import ast
from enum import Enum
from .verdicts.ExceptionVerdict import ExceptionVerdict
from .verdicts.FailedVerdict import FailedVerdict
from .verdicts.ExampleVerdict import ExampleVerdict
from ..properties import PENDING_STATE
from ..utils import create_node_representation, send_current_state
from .doctest_parser import DocTestParser, Example
from typing import List, Tuple

# Only these types can have a docstring 
# according to ast's module documentation
SUPPORTED_TYPES = (ast.FunctionDef, 
                   ast.ClassDef, 
                   ast.Module)

class L1DocTestFlag(Enum):
    """This enum represents the flag of an L1DocTest. A flag is an integer that 
    represents the global verdict of the L1DocTest. 
    
    The flag can be one of the following: `FAILED_FLAG`, `EMPTY_FLAG`, `PASSED_FLAG`.
    """
    FAILED_FLAG = -1
    EMPTY_FLAG = 0
    PASSED_FLAG = 1
    
    def from_value(value: int):
        """This method returns the flag corresponding to the given value."""
        if value not in [-1, 0, 1]:
            raise ValueError("The value must be one of the following: -1, 0, 1")
        if value == -1:
            return L1DocTestFlag.FAILED_FLAG
        elif value == 0:
            return L1DocTestFlag.EMPTY_FLAG
        else:
            return L1DocTestFlag.PASSED_FLAG
    
    def get_image_basename(self):
        """This method returns the icon corresponding to the flag."""
        if self.value == -1:
            return "failed.png"
        elif self.value == 0:
            return "warning.png"
        else:
            return "passed.png"
    
    def get_color(self):
        """This method returns the color corresponding to the flag."""
        if self.value == -1:
            return "red"
        elif self.value == 0:
            return "orange"
        else:
            return "green"
        
    def short_name(self):
        """This method returns the text corresponding to the flag."""
        if self.value == -1:
            return "Failed"
        elif self.value == 0:
            return "Empty"
        else:
            return "Passed"
    
    @staticmethod
    def get_priorities():
        return [L1DocTestFlag.FAILED_FLAG, L1DocTestFlag.PASSED_FLAG, L1DocTestFlag.EMPTY_FLAG]
    
    def is_failing(self):
        """
        Return True if the flag is `FAILED_FLAG`. Otherwise, returns False.
        """
        return self.value == -1 
    
    
class L1DocTest():
    """
    An `L1DocTest` is associated to an ast node and that groups its tests 
    containing in the docstring of that node. A `L1DocTest` handles a list of 
    the examples of type `Example`. 
    
    When a `L1DocTest` is created by the `L1TestAstParser` it is not evaluated. 
    Its global verdict (see `get_flag()`) will be `None` until it's evaluated. 
    Its `Example` are not evaluated too and thier verdicts are setted as `None`. 
    To evaluate a `L1DocTest` you should call the `evaluate()` method. Then all 
    the examples will be evaluated and the global verdict of the l1doctest will 
    be setted.
    
    Args:
        filename (str): The name of the file in which the L1Doctest is extracted.
        This is useful to know the context of the L1Doctest.
        name (str):  The name of the ast node which is associated to the L1Doctest.
        type (str): The type of the ast node which is associated to the L1Doctest. 
        related to one value reported in `SUPPORTED_TYPES` constant.
        node_lineno (int): The line of the associated ast node.
        start_lineno (int): the starting lineno of the docstring.
        Defaults to -1 If no docstring was found.
        end_lineno (int): the ending lineno of the docstring.
        Defaults to -1 If no docstring was found.
    """
    def __init__(self, filename:str, name:str, type:str, node_lineno:int, start_lineno=-1, end_lineno=-1) -> None:
        self.__filename = filename
        self.__name = name
        self.__type = type
        self.__node_lineno = node_lineno
        self.__start_lineno = start_lineno
        self.__end_lineno = end_lineno
        self.__examples:List[Example] = []
        self.__flag: L1DocTestFlag = None  # The global verdict of the l1doctest. 
                                           # Set as None while the l1doctest is not evaluated.
    
    def evaluate(self, globs:dict):
        """
        Evaluate the examples of this L1Doctest. Each example is evaluated and its 
        verdict is setted. The global verdict of the L1Doctest is also setted and you
        could access it by calling the method L1Doctest.get_flag().
        
        Args:
            globs (dict): The globals of the module in wich the L1Doctest is defined.
        Returns:
            L1DocTestFlag: The global verdict of the L1Doctest.
        """
        for example in self.get_examples(): 
            send_current_state(state=PENDING_STATE, lineno=example.lineno)
            example.compute_and_set_verdict(globs) 
        self.__flag = self.__get_global_flag() # set the global verdict of the l1doctest 
        return self.__flag
        
    def __get_global_flag(self) -> L1DocTestFlag:
        """
        This method should be invoked after the evaluation (evaluate_examples() method) 
        of the l1doctest. It returns a flag that represents the global verdict of the L1Doctest.
        
        Returns:
            L1DocTestFlag: returns a flag that represents the global 
            verdict of the l1doctest. 
        """
        assert all(ex.is_evaluated() for ex in self.__examples) # assure that all the examples are evaluated
        if not self.has_examples():
            return L1DocTestFlag.EMPTY_FLAG
        elif self.has_only_setUps() and all([ex.get_verdict().isSuccess() for ex in self.__examples]):
            return L1DocTestFlag.EMPTY_FLAG
        elif all([ex.get_verdict().isSuccess() for ex in self.__examples]):
            return L1DocTestFlag.PASSED_FLAG
        else:
            return L1DocTestFlag.FAILED_FLAG
    
    ## Stats methods ##
    def count_tests(self):
        """
        Count the number of the tests containing in this L1doctest. The setups
        are not considered in the counting.
        
        Returns:
            int: the number of the tests containing in the l1doctest.
        """ 
        return sum([1 for ex in self.__examples if ex.is_a_test()])
    
    def count_setups(self):
        """
        Count the number of the setups containing in this L1doctest. The tests
        are not considered in the counting.
        
        Returns:
            int: the number of the setups containing in the l1doctest.
        """ 
        return sum([1 for ex in self.__examples if not ex.is_a_test()])
    
    def count_passed_tests(self):
        """Count the number of passed tests. Return `None` if the L1Doctest is not evaluated."""
        if not self.is_evaluated():
            return None
        return sum([1 for ex in self.__examples if ex.is_a_test() and ex.get_verdict().isSuccess()])
    
    def count_failed_tests(self):
        """Count the number of failed tests. Return `None` if the L1Doctest is not evaluated."""
        if not self.is_evaluated():
            return None
        return sum([1 for ex in self.__examples if not ex.get_verdict().isSuccess() and isinstance(ex.get_verdict(), FailedVerdict)])
    
    def count_error_tests(self):
        """
        Count the number of tests that raised an exception.
        Return `None` if the L1Doctest is not evaluated
        """
        if not self.is_evaluated():
            return None
        return sum([1 for ex in self.__examples if not ex.get_verdict().isSuccess() and isinstance(ex.get_verdict(), ExceptionVerdict)])
    
    ## Sorting/filtering methods ##
    def sort_examples_by_verdicts(self, key: Tuple[ExampleVerdict, ...], reverse=True):
        """
        Sort the examples of this L1Doctest by the given key. The key is a tuple of `ExampleVerdict`.
        This method set the examples of this L1Doctest as the sorted examples.
        """
        sorted_example = sorted(self.get_examples(), key=lambda x: isinstance(x.get_verdict(), key), reverse=reverse)
        self.set_examples(sorted_example)
    
    def filter_examples_by_verdicts(self, key: Tuple[ExampleVerdict, ...]):
        """Filter the examples of this L1Doctest by the given key. The key is a tuple of `ExampleVerdict`.
        Return a list of examples that match the given key.
        """
        return [example for example in self.__examples if example.get_verdict().__class__ in key]
    
    ## Check methods ##
    def is_evaluated(self):
        """Returns True if the L1Doctest is evaluated. Otherwise, returns False."""
        return bool(self.__flag)
    
    def has_examples(self):
        """
        Returns:
            bool: Returns True if this `L1Doctest` contains at least one `Example`. 
            Otherwise, returns False.
        """
        return len(self.__examples) > 0
    
    def has_only_tests(self):
        """ 
        Return True if this L1Doctest contains only tests. Otherwise, returns False.
        """
        return self.count_tests() > 0 and self.count_tests() == len(self.__examples)
    
    def has_only_setUps(self):
        """
        Return True if this L1Doctest contains only setups. Otherwise, returns False.
        """
        return self.count_setups() > 0 and self.count_setups() == len(self.__examples)
    
    ## Add and remove methods ##
    def add_example(self, example:Example):
        """
        Add an example to the L1Doctest.

        Args:
            example (Example): The example to add.
        """
        self.__examples.append(example)
        
    def remove_example(self, example:Example):
        """
        Remove an example from the L1Doctest.

        Args:
            example (Example): The example to remove.
        """
        self.__examples.remove(example)
    
    ## Utils methods ##
    def get_verdict_from_example(self, example:Example) -> ExampleVerdict:
        """
        Return the verdict of the given example. Return None if the example is not found.
        """
        found_example = self.__get_example(example.lineno)
        return found_example.get_verdict() if found_example else None
    
    def get_verdict_from_lineno(self, lineno:int) -> ExampleVerdict:
        """
        Return the verdict of the example that has the given lineno. 
        Return None if the example is not found.
        """
        found_example = self.__get_example(lineno)
        return found_example.get_verdict() if found_example else None
    
    def get_example_from_lineno(self, lineno:int) -> Example:
        """
        Return the example that has the given lineno. Return None if the example is not found.
        """
        return self.__get_example(lineno)
    
    def __get_example(self, lineno:int) -> Example:
        """
        Return the example that has the given lineno. 
        Return None if the example is not found.
        """
        found = [ex for ex in self.__examples if ex.lineno == lineno]
        return found[0] if found else None

    ## Getters and Setters ##
    def get_filename(self):
        return self.__filename
    
    def get_node_lineno(self):
        """Return the line number of the node of the L1Doctest. 
        The node is the function or the class associated to the L1Doctest. """
        return self.__node_lineno
    
    def get_examples(self):
        return self.__examples
    
    def set_examples(self, examples:List[Example]):
        self.__examples = examples
    
    def get_name(self):
        """Return the name of the L1Doctest. The name is the name of the function or the class."""
        return self.__name
    
    def get_start_end_lineno(self) -> Tuple[int, int]:
        """Return the start and the end line number of the docstring of the L1Doctest."""
        return (self.__start_lineno, self.__end_lineno)

    def get_start_lineno(self):
        """Return the start line number of the docstring of the L1Doctest.""" 
        return self.__start_lineno
    
    def set_start_lineno(self, start_lineno):
        self.__start_lineno = start_lineno
    
    def get_end_lineno(self):
        """Return the end line number of the docstring of the L1Doctest."""
        return self.__end_lineno
    
    def set_end_lineno(self, end_lineno):
        self.__end_lineno = end_lineno
        
    def get_flag(self):
        """Return the flag of the L1Doctest. The flag is the global verdict of the L1Doctest."""
        return self.__flag
    
    def get_type(self):
        """Return the type of the L1Doctest. The type is the type of the node of the L1Doctest."""
        return self.__type
    
    def set_type(self, type):
        self.__type = type


## The L1DocTestParser class 
class L1TestAstParser():
    """
    The L1DocTestParser is responsible for parsing the source and extracts the L1DocTests.
    The main method is `parse()` that returns a generator of L1DocTests. The generator 
    yields all the L1DocTests found in the source. 
    
    Note: generators are used to avoid to load all the L1DocTests in memory. This
    is useful when the source is big. The L1DocTests are yielded one by one while parsing.
    Finally, the use of generators allows better performances and less memory consumption.
    
    The `L1DocTestParser`doesn't evaluate the created L1Doctests. Each yielded L1DocTest
    is initially unevaluated. It's the responsibility of the `Evaluator` to evaluate each 
    received L1DocTest.
    
    Example of use:
    >>> parser = L1DocTestParser("source", "filename.py")
    >>> generator:  = parser.parse()
    >>> 
    """
    def __init__(self, source:str="", filename:str="", mode="exec") -> None:
        self._filename = filename
        self._source = source
        self._mode = mode

    def parse(self, lineno:int=None):
        """
        Parse the source and yields all the l1doctest.
        If the line number is not None, only the l1doctest that has 
        the given line number is yielded.
        
        Args:
            lineno (int): the line number of the l1doctest to yield. If None, 
            all the l1doctests are yielded.
        Returns:
            generator: a generator of l1doctests.
        
        Raises: 
            Error: A compilation error raised by the AST module.
        """
        body = self.__get_ast_body(self._source)
        yield from self.__recursive_walking(body, lineno)
    
    def __get_ast_body(self, source) -> List[ast.AST]:
        """
        Returns the body(list of nodes) of the root node of the AST tree.
        
        Raises: 
            Error: A compilation error raised by the AST module.
        """
        return ast.parse(source, self._filename, mode=self._mode).body
    
    def __recursive_walking(self, list_nodes: list, lineno: int = None):
        """
        Search all the supported nodes in the AST tree. Even sub-nodes are visited.
        This is a recursive function, so all the visited nodes are added in the ~all_nodes~
        parameter.
        """
        doctest_parser = DocTestParser()
        for node in list_nodes:
            if isinstance(node, SUPPORTED_TYPES):
                type = node.__class__.__name__
                l1doctest = L1DocTest(self._filename, create_node_representation(node), type, node.lineno)
                if node.body:
                    first_stmt = node.body[0]
                    if isinstance(first_stmt, ast.Expr) and isinstance(first_stmt.value, ast.Str):
                        l1doctest.set_start_lineno(first_stmt.lineno)
                        l1doctest.set_end_lineno(first_stmt.end_lineno)

                docstring = ast.get_docstring(node, False) or ""
                parsed_doc = doctest_parser.parse(docstring, name=self._filename)

                l1doctest.set_examples(self.__filter_example_instances(l1doctest, parsed_doc))
                # Check if the current node matches the given lineno
                if lineno is None or (l1doctest.get_node_lineno() == lineno):
                    yield l1doctest

                yield from self.__recursive_walking(ast.iter_child_nodes(node), lineno)

    def __filter_example_instances(self, l1_doctest:L1DocTest, parsed_doc:list) -> List[Example]:
        """
        The filter can return an empty list if there's no example 
        in the parsed docstring. Otherwise, it will return a list of the `Example` instances 
        matching the given l1doctest.
        """
        examples = []
        for test in parsed_doc:
            if isinstance(test, Example):
                test.lineno = test.lineno + l1_doctest.get_start_lineno() - 1
                examples.append(test)
        return examples    
        
    def get_filename(self):
        return self._filename
    
    def set_filename(self, filename: str):
        self._filename = filename
    
    def get_source(self):
        return self._source
    
    def set_source(self, source: str):
        self._source = source
        
    def get_mode(self):
        return self._mode
    
    def set_mode(self, mode):
        self._mode = mode



from .ExampleVerdict import ExampleVerdict

class PassedSetupVerdict(ExampleVerdict):
    def __init__(self, filename, lineno, tested_line):
        super().__init__(filename, lineno, tested_line, "None")

    def get_color(self):
        return "black"

    def isSuccess(self):
        return True
    
    def _verdict_message(self):
        return "Execution succeed"


from thonny import  tktextext, ui_utils
from thonny.ui_utils import scrollbar_style
from ..backend.verdicts.PassedVerdict import PassedVerdict
from ..backend.verdicts.PassedSetupVerdict import PassedSetupVerdict
from thonnycontrib.utils import get_font_size_option, get_font_family_option
from .l1test_error_view import L1TestErrorView
from ..properties import *
from ..backend.doctest_parser import Example
from ..backend.ast_parser import L1DocTest, L1DocTestFlag
from ..backend.verdicts.ExceptionVerdict import ExceptionVerdict
from ..backend.verdicts.FailedVerdict import FailedVerdict
from ..backend.verdicts.FailedWhenExceptionExpectedVerdict import FailedWhenExceptionExpectedVerdict
from ..utils import *
from functools import partial
from ..l1test_configuration.l1test_options import *
import tkinter as tk, tkinter.font as tk_font, thonny
from collections import namedtuple
from thonny.codeview import *
from typing import Dict, List
from copy import deepcopy
from tkinter import ttk

# La hauteur, par défault, d'une ligne dans une Treeview
ROW_HEIGHT = 40

SMALL_MARGIN = 1.1
NORMAL_MARGIN = 1.17

CLICKABLE_TAG = "clickable"

# L'objet qui représente un summarize
Summarize = namedtuple('Summarize', ["total", "success", "failures", "errors", "empty"])

# Palette de couleurs utilisée par la treeview
COLORS:dict = { 'orange': '#e8770e',
                'red': 'red',
                'lightred': '#ffdddb',
                'darkred': '#f7140c',
                'green': 'darkgreen',
                'blue': '#0000cc',
                'gray': 'gray'
            }

## Binding events to shortcuts.  
ALT_I = "<Alt-i>" # to increase space between tree rows
ALT_D = "<Alt-d>" # to decrease space between tree rows
ALT_F = "<Alt-f>" # to update font size
ALT_U = "<Alt-u>" # to unfold tree rows
ALT_O = "<Alt-o>" # to fold tree rows
BINDINGS = {ALT_I: "Alt+i", ALT_D: "Alt+d", ALT_F: "Alt+f", ALT_U: "Alt+u", ALT_O: "Alt+o"}

class L1TestTreeView(ttk.Frame):    
    def __init__(self, master=None):
        ttk.Frame.__init__(self, master, borderwidth=0, relief="flat")
        self.workbench = thonny.get_workbench()
        self._origin_l1doctests: List[L1DocTest] = dict()
        self._copy_l1doctests = self._origin_l1doctests.copy()
        
        self.__init_treeview()
        self.__init_workbench_bindings()
        self.__init_special_attributes()
    
    def __init_workbench_bindings(self):
        """
        Binds the events to the workbench.
        """
        # Le binding est censé augmenter/diminuer automatiquement la taille de la treeview
        self.workbench.bind("<Control-plus>", self.observe_font_changing, True)
        self.workbench.bind("<Control-minus>", self.observe_font_changing, True)
        self.workbench.bind(ALT_I, self.increase_row_height, True)
        self.workbench.bind(ALT_D, self.decrease_row_height, True)
        self.workbench.bind(ALT_F, self.update_font, True)
        self.workbench.bind(ALT_U, self.expand_rows, True) 
        self.workbench.bind(ALT_O, self.fold_rows, True)
               
    def __init_treeview(self):
        """
        Creates the treeview widget.
        """
        self.vert_scrollbar = ttk.Scrollbar(self, orient=tk.VERTICAL, style=scrollbar_style("Vertical"))
        self.vert_scrollbar.grid(row=0, column=1, sticky=tk.NSEW, rowspan=3)
    
        self.treeview = ttk.Treeview(self,yscrollcommand=self.vert_scrollbar.set)
        rows, columns = 2, 0
        self.treeview.grid(row=rows, column=columns, sticky=tk.NSEW)
        self.vert_scrollbar["command"] = self.treeview.yview
        self.columnconfigure(columns, weight=1)
        self.rowconfigure(rows, weight=1)
        self.treeview.column("#0", anchor=tk.W, stretch=tk.YES) # configure the only tree column
        self.treeview["show"] = ("tree",)

        for color_name, color in COLORS.items():
            if color_name == "lightred": # lightred is used for background (highlighting)
                self.treeview.tag_configure(color_name, background=color)
            else:
                self.treeview.tag_configure(color_name, foreground=color)
        
        # définir un tag pour les tests en exception: il faut les souligner (underline)
        self.treeview.tag_configure(
            "exception_as_link", 
            foreground=COLORS["red"], 
            font=tk_font.Font(underline=True, weight="normal", family=get_font_family_option(), size=get_font_size_option())
        )
        # définir un tag pour les tests en exception survolé: il faut les souligner (underline) et les mettre en gras (bold)
        self.treeview.tag_configure(
            "exception_hovered", 
            foreground=COLORS["red"], 
            font=tk_font.Font(underline=True, weight="bold", family=get_font_family_option(), size=get_font_size_option())
        )

        # définir un style par défault pour la treeview
        self.style = ttk.Style()
        self.style_mapping = self.style.map('Treeview')
        self.__update_tree_font(get_font_family_option(), get_font_size_option())

        # All the icons used in the treeview should be loaded here to keep thier references.
        # Otherwise, they will be garbage collected and the treeview will not show them.
        self.image_references = {flag.get_image_basename(): get_photoImage(flag.get_image_basename()) for flag in L1DocTestFlag}
        self.image_references[PENDING_ICON] = get_photoImage(PENDING_ICON)
        self.image_references[ERROR_ICON] = get_photoImage(ERROR_ICON)
        self.image_references[RESTART_ICON] = get_photoImage(RESTART_ICON)
        self.image_references[FAILED_RED_CHIP] = get_photoImage(FAILED_RED_CHIP)
        self.image_references[EXCEPTION_RED_CHIP] = get_photoImage(EXCEPTION_RED_CHIP)
        self.image_references[SUCCESS_GREEN_CHIP] = get_photoImage(SUCCESS_GREEN_CHIP)
        
        self.treeview.tag_bind("clickable", "<<TreeviewSelect>>", self._on_select)
        self.treeview.tag_bind("nonClickable", "<<TreeviewSelect>>", self._remove_highlight_selection_effect)
        self.treeview.bind("<Motion>", self._on_hover_exception_test)
        self.treeview.bind("<Configure>", self.__wrap_tree_content) # Here we handle the motion event of the treeview 
        
        # add a menu to the treeview 
        self.menu = tk.Menu(self.treeview, name="menu", tearoff=False)        
        self.init_header(row=0, column=0)
        
    def __init_special_attributes(self):
        """
        Initializes the special attributes of the treeview.
        """
        self.__old_width = -1 # utilisé pour savoir si la hauteur de la treeview a changé
        self.__max_lines = 1 # le nombre de lignes du texte le plus long de la treeview
        self.__hovered_exception = None # l'exception test sur laquelle le curseur est en train de passer
     
    def _on_hover_exception_test(self, event):
        """
        Handles the motion event of the treeview. When the cursor is hovering over an exception test, 
        the font of the test is changed to bold and underlined. The cursor is also changed to "hand".
        """
        if not self.is_empty():
            item = self.treeview.identify('item', event.x, event.y)
            tags = self.treeview.item(item, "tags")
            if "exception_as_link" in tags:
                if self.__hovered_exception and self.__hovered_exception != item:
                    # If the cursor is hovering over a different item, revert the font and cursor to normal for the previously hovered item
                    prev_item = self.__hovered_exception
                    prev_tags = self.treeview.item(prev_item, "tags")
                    new_tags = tuple(tag for tag in prev_tags if tag not in ("exception_hovered", "exception_cursor"))
                    self.treeview.item(prev_item, tags=new_tags + ("exception_as_link",))
                    self.treeview.configure(cursor="")
                    self.__hovered_exception = None

                # Change the font to bold and underlined and set the cursor to "hand" when the cursor enters the row with the "exception_as_link" tag
                new_tags = tuple(tag for tag in tags if tag != "exception_as_link")
                self.treeview.item(item, tags=new_tags + ("exception_hovered", "exception_cursor"))
                self.treeview.configure(cursor="hand2")
                self.__hovered_exception = item
            elif self.__hovered_exception and self.__hovered_exception != item:
                # If the cursor is not over any "exception_as_link" item, revert the font and cursor to normal for the previously hovered item
                prev_item = self.__hovered_exception
                prev_tags = self.treeview.item(prev_item, "tags")
                new_tags = tuple(tag for tag in prev_tags if tag not in ("exception_hovered", "exception_cursor"))
                self.treeview.item(prev_item, tags=new_tags + ("exception_as_link",))
                self.treeview.configure(cursor="")
                self.__hovered_exception = None

    # ------------------------------------
    # Fonctions pour le Header du treeview
    # ------------------------------------
    
    def init_header(self, row=0, column=0):
        """
        Initialize the header of the treeview. Initially, the header contains 
        only a `menu button` at the top right of the treeview. The options of the 
        menu are created by `post_button_menu()` method.

        Args:
            row (int): Always set to 0. Defaults to 0.
            column (int): Always set to 0. Defaults to 0.
        """
        header_frame = ttk.Frame(self, style="ViewToolbar.TFrame")
        header_frame.grid(row=row, column=column, sticky="nsew")
        header_frame.columnconfigure(0, weight=1)
        
        self.header_bar = tktextext.TweakableText(
            header_frame,
            borderwidth=0,
            relief="flat",
            height=1,
            wrap="word",
            padx=ui_utils.ems_to_pixels(0.6),
            pady=ui_utils.ems_to_pixels(0.5),
            insertwidth=0,
            highlightthickness=0,
            background=ui_utils.lookup_style_option("ViewToolbar.TFrame", "background"),
        )
        
        for color_name, color in COLORS.items():
            self.header_bar.tag_configure(color_name, foreground=color)
        
        self.header_bar.grid(row=0, column=0, sticky="nsew")
        self.header_bar.set_read_only(True)
        self.menu_button = ttk.Button(
                                header_frame, 
                                text=" ≡ ", 
                                style="ViewToolbar.Toolbutton", 
                                width=3,
                                command=self.post_button_menu
                            )
        self.header_bar.bind("<Configure>", self.resize_header_bar, True)
        self.header_bar.configure(height=1)

        self.menu_button.place(anchor="ne", rely=0, relx=1)
        self.disable_menu()
     
    def post_button_menu(self):
        """
            The handler of the menu button located in the header of the treeview.
            When clicking the menu button a popoup is opened and shows several options.
        """
        self.add_menu_options()
        self.menu.tk_popup(
            self.menu_button.winfo_rootx(),
            self.menu_button.winfo_rooty() + self.menu_button.winfo_height(),
        )
    
    def add_menu_options(self):
        """
            Adds the options inside the menu button.
        """
        self.menu.delete(0, "end") 
         
        self.menu.add_command(label=PLACE_RED_TEST_ON_TOP_LABEL, command=self.sort_by_red_tests)
        self.menu.add_command(label=RESUME_ORIGINAL_ORDER, command=self.remove_sort_filter)
        self.menu.add_separator() 
        self.menu.add_command(label=SHOW_ONLY_RED_TESTS, command=self.show_only_red_tests)
        self.menu.add_command(label=SHOW_ALL_TESTS, command=self.show_all_tests)
        self.menu.add_separator()
        self.menu.add_command(label=GROUP_BY_VERDICTS, command=self.group_by_verdicts)
        self.menu.add_separator()    
        self.menu.add_command(label=EXPAND_ALL, command=self.expand_rows, accelerator=BINDINGS[ALT_U]) 
        self.menu.add_command(label=FOLD_ALL, command=self.fold_rows, accelerator=BINDINGS[ALT_O])
        self.menu.add_separator() 
        self.menu.add_command(label=UPDATE_FONT_LABEL, command=self.update_font, accelerator=BINDINGS[ALT_F])
        self.menu.add_command(label=INCREASE_SPACE_BETWEEN_ROWS, command=self.increase_row_height, accelerator=BINDINGS[ALT_I])
        self.menu.add_command(label=DECREASE_SPACE_BETWEEN_ROWS, command=self.decrease_row_height, accelerator=BINDINGS[ALT_D])
        self.menu.add_separator()
        self.menu.add_command(label=CLEAR_LABEL, command=partial(self.clear_tree, clear_verdicts_data=True))
          
    def resize_header_bar(self, event=None):
        """ 
        Resize the height of the header. 
        Always keep this method otherwise the header will take the whole treeview.
        """
        height = self.tk.call((self.header_bar, "count", "-update", "-displaylines", "1.0", "end"))
        self.header_bar.configure(height=height)

    def update_font(self, event=None):
        """
        This is the handler of the `Update the font` option.  
        
        It also update the font size of the treeview if and 
        only if the font is changed in thonny.
        """          
        self.resize_header_bar()
        # on applique la nouvelle police pour la treeview
        self.observe_font_changing()
     
    def sort_by_red_tests(self):
        
        # On filtre que les l1doctest qui ont un flag rouge.
        sorted_failed_nodes = [l1doctest for l1doctest in self._copy_l1doctests if l1doctest.get_flag() == L1DocTestFlag.FAILED_FLAG ]
        
        # On met les tests rouges en haut et cela pour tous les l1doctest qui ont au moins un test rouge.
        sorted_failed_tests = []
        for l1doctest in sorted_failed_nodes:
            new_l1doctest = deepcopy(l1doctest)
            new_l1doctest.sort_examples_by_verdicts((FailedVerdict, ExceptionVerdict))
            sorted_failed_tests.append(new_l1doctest)
           
        # On récupère les autres neouds AST 
        others = [l1doctest for l1doctest in self._copy_l1doctests if l1doctest.get_flag() != L1DocTestFlag.FAILED_FLAG]
         # l'addition de tous les noeuds
        sorted_failed_tests.extend(others)

        if not self.is_empty(): # on met à jour que si la treeview possède des éléments
            self._copy_l1doctests = sorted_failed_tests
            self.update_tree_contents(sorted_failed_tests, clear_header=False, clear_errorview=False)
    
    def remove_sort_filter(self):
        """
        When invoking this method the treeview will restore 
        the original order of its rows.
        """
        def __is_verdicts_copy_equals_to_original():
            if len(self._copy_l1doctests) != len(self._origin_l1doctests):
                return False
            c_names = [c_l1doctest.get_name() for c_l1doctest in self._copy_l1doctests]
            for o_l1doctest in self._origin_l1doctests:
                c_index = c_names.index(o_l1doctest.get_name())
                c_l1doctest = self._copy_l1doctests[c_index]
                if len(c_l1doctest.get_examples()) != len(o_l1doctest.get_examples()): 
                    return False 
            return True
        if not self.is_empty(): # on met à jour que si la treeview possède des éléments
            if __is_verdicts_copy_equals_to_original():
                self._copy_l1doctests = self._origin_l1doctests
            self.update_tree_contents(self._copy_l1doctests, clear_header=False, clear_errorview=False)
    
    def show_only_red_tests(self):
        from copy import deepcopy
        red_verdicts = (FailedVerdict, FailedWhenExceptionExpectedVerdict, ExceptionVerdict)
        
        only_reds = []
        for l1doctest in self._copy_l1doctests:
            if l1doctest.get_flag() == L1DocTestFlag.FAILED_FLAG:
                red_examples = l1doctest.filter_examples_by_verdicts(red_verdicts)
                new_l1doctest = deepcopy(l1doctest)
                new_l1doctest.set_examples(red_examples)
                only_reds.append(new_l1doctest)
                            
        if not self.is_empty(): # on met à jour que si la treeview possède des éléments
            self._copy_l1doctests = only_reds if only_reds else self._copy_l1doctests
            self.update_tree_contents(self._copy_l1doctests, clear_header=False, clear_errorview=False)
       
    def show_all_tests(self):
        self._copy_l1doctests = self._origin_l1doctests
        self.update_tree_contents(self._copy_l1doctests, clear_header=False, clear_errorview=False)
        
    
    def __group_by_verdicts(self, l1doctests: List[L1DocTest], status_order: List[L1DocTestFlag]=None) -> Dict[L1DocTestFlag, List[L1DocTest]]: 
        """
        Group l1doctests by status and order the statuses according to the provided 
        status order.
        
        :param status_order: A list of L1DocTestFlag that specifies the order in which 
        to group the verdicts by flag.
        :return: A dictionary of verdicts grouped by status.
        """
        status_groups:Dict[L1DocTestFlag, List[L1DocTest]] = {}
        for l1doctest in l1doctests:
            status = l1doctest.get_flag()
            if status not in status_groups:
                status_groups[status] = []
            status_groups[status].append(l1doctest)

        if status_order is None:
            status_order = L1DocTestFlag.get_priorities()
        
        # mettre en ordre les catégories selon l'ordre défini par le paramètre `status_order`
        ordered_groups = {}
        for status in status_order:
            if status in status_groups:
                ordered_groups[status] = status_groups[status]
        status_groups = ordered_groups
        
        return status_groups
    
    def group_by_verdicts(self):  
        grouped = self.__group_by_verdicts(self._origin_l1doctests)
        if not self.is_empty(): # on met à jour que si la treeview possède des éléments
            self.clear_tree(clear_all=False) 
            for status, l1doctests in grouped.items():
                row_id = self.treeview.insert(parent='', index="end", text=status.short_name(), open=status.is_failing(), 
                                         tags=(status.get_color(),))
                self.__add_verdicts_to_treeview(l1doctests, row_id)
        
    def expand_rows(self, event=None): 
        """ Spreads the function rows and the red tests. """
        self.__apply_row_change(False, get_option(OPEN_RED_TEST_ROWS))
    
    def fold_rows(self, event=None):
        """ Folds only the function rows """
        self.__apply_row_change(True, get_option(OPEN_RED_TEST_ROWS)) 
    
    def __apply_row_change(self, close_func_rows:bool, expand_red_tests:bool):
        """
        This function is used to expand or fold the rows of the treeview.
        It applies the new values of the args `close_rows` and `expand_red_tests` to
        the options `CLOSE_ALL_ROWS` and `OPEN_RED_ROWS` respectively. Then, it calls
        the method `update_tree_contents()` to update the treeview.

        Args:
            close_func_rows (bool): If True, the function rows will be folded. 
            expand_red_tests (bool): If True, the red tests will be expanded.
        """
        if not self.is_empty(): # on met à jour que si la treeview possède des éléments
            from copy import copy
            # on sauvegarde les valeurs initiales des options
            orig_red_tests = copy(get_option(OPEN_RED_TEST_ROWS))
            orig_close_rows = copy(get_option(CLOSE_FUNCTION_ROWS))
            
            # on applique les nouvelles valeurs aux options
            set_option(OPEN_RED_TEST_ROWS, expand_red_tests)
            set_option(CLOSE_FUNCTION_ROWS, close_func_rows)
            
            # update_tree invoque dynamiquement la valeur des options
            self.update_tree_contents(self._copy_l1doctests, clear_header=False, clear_errorview=False)
            
            # on restaure les valeurs initiales des options
            set_option(OPEN_RED_TEST_ROWS, orig_red_tests)
            set_option(CLOSE_FUNCTION_ROWS, orig_close_rows)
    
    def increase_row_height(self, event=None):
        if not self.is_empty():
            current_height = self.get_current_rowheight()
            self.update_row_height(current_height+1)
    
    def decrease_row_height(self, event=None):
        if not self.is_empty():
            current_height = self.get_current_rowheight()
            max_lines = self.get_max_lines()
            opt = self.__compute_optimal_height(max_lines, SMALL_MARGIN)
            if current_height > opt:
                self.update_row_height(current_height-1) 

    def update_row_height(self, rowheight:int=None):
        """
        Updates the height of a row in the treeview.
        """
        rowheight = rowheight if rowheight else self.get_current_rowheight()
        self.style.configure("Treeview", rowheight=rowheight) 
    
    def get_current_rowheight(self):
        """
        Returns the current height of a row in the treeview.
        """
        return self.style.lookup("Treeview", 'rowheight')    
    # -------------------------------------------
    # Fin des fonction pour le Header du treeview
    # ------------------------------------------
    
    def observe_font_changing(self, event=None):
        """
        Changes the font of the treeview.
        """
        self.__update_tree_font(get_font_family_option(), 
                                get_font_size_option())
        self.change_header_font() 
        self.__wrap_tree_content()
        error_view:L1TestErrorView = self.workbench.get_view(L1TestErrorView.__name__)
        error_view.update_font()
     
    def change_header_font(self, header_font_size=11):
        self.header_bar.config(font=(get_font_family_option(), header_font_size))
        self.resize_header_bar()
    
    def __update_tree_font(self, font_family, font_size):
        """
        Applies the new font to the treeview.
        """
        self.style.configure("Treeview", justify="left", font=(font_family, font_size), wrap=tk.WORD)  
        self.treeview.tag_configure("exception_as_link", font=tk_font.Font(underline=True, family=font_family, size=font_size))
    
    def __wrap_tree_content(self, event=None, margin=NORMAL_MARGIN):
        """
            This function wraps the text of treeview to follow its width.
        """
        widget = event.widget if event else self.treeview
        if (isinstance(widget, ttk.Treeview)): 
            view = self.workbench.get_view(self.__class__.__name__)
            if view.winfo_ismapped() :  
                if not self.is_empty():
                    width = view.winfo_width()
                    if self.__width_is_changed(width): # on ne met à jour que si la hauteur de la treeview n'a pas changé        
                        chars_per_pixels = width // get_font_size_option() 
                        visible_nodes = self.get_all_tree_childrens()
                        longest_length = self.__update_wrapped_texts(visible_nodes, chars_per_pixels)
                        new_rowheight = self.__compute_optimal_height(longest_length, margin)
                        self.update_row_height(new_rowheight)
                    self.__old_width = width
    
    def __width_is_changed(self, width):
        """
        Returns True if the height of the treeview hasn't changed.
        """
        return self.__old_width != width
    
    def get_all_tree_childrens(self):
        """
        Gets recursivly all the childrens of the given node of the treeview. 
        
        Note: This function is a generator. It yields the childrens of the given node 
        one by one. This is useful to not load all the childrens in memory at once after
        each update of the treeview.
        
        Keep this method as a generator to avoid performance issues on th GUI. 
        It's observed that thonny's IDE is more faster when the treeview is 
        updated with a generator.
        """
        def _all_childrens(treeview: ttk.Treeview, node: str = None):
            child = treeview.get_children(node)
            for sub_child in child: 
                if treeview.item(sub_child)["open"]:
                    yield from _all_childrens(self.treeview, sub_child)
                yield sub_child  

        yield from _all_childrens(self.treeview)

    def __update_wrapped_texts(self, nodes, chars_per_pixels:int) -> int:
        """
        Met à jour le wrapping des textes des nœuds spécifiés dans la Treeview.
        Return the longest wrapped line of the treeview.
        Returns the number of lines of the longest wrapped line of the treeview.
        """
        max_lines = 1
        for node in nodes:
            text = self.treeview.item(node, "text")
            wrapped_text = wrap(text, chars_per_pixels)
            
            num_lines = wrapped_text.count("\n") + 1
            if num_lines > max_lines:
                max_lines = num_lines
                
            self.treeview.item(node, text=wrapped_text)
        self.__max_lines = max_lines
        return self.__max_lines
    
    def __compute_optimal_height(self, max_lines:int=None, margin=NORMAL_MARGIN):
        """
            Uses the default font metrics to calculate the optimal row height.
            The default font metrics is multiplied by the given `max_lines`.
            
            Args:
                max_lines(int): The number of lines of the longest row in the treeview.
            Return:
                (int): The new height.
        """
        row_height = get_font_size_option() * 2     # multiply by 2 to handle the line spacing
        opt_height = max_lines * (row_height * margin) if max_lines else row_height
        return round(opt_height)
    
    def update_tree_contents(self, l1doctests:List[L1DocTest], parent="", clear_errorview=True, clear_header=True):
        """
            This function contructs and inserts the rows into the treeview.
        """
        self.__old_width = self.workbench.get_view("L1TestTreeView", False).winfo_width() if self.__old_width > 0 else -1
        
        self._restore_row_selection_effect() 
        self.clear_tree(clear_all=clear_header, clear_errorview=clear_errorview)
    
        if not self.__check_if_editor_is_open():
            return
        
        if not l1doctests:
            self.insert_in_header("No test found !", image="warning.png", clear=True, tags=("orange",))
            return
        
        self.enable_menu()
        self.__add_verdicts_to_treeview(l1doctests, parent)
        
        # on insère le summarize dans le header bar que si la treeview n'est pas vide
        if not self.is_empty() and clear_header:
            # We build the summarize object 
            summarize: Summarize = self.build_summarize_object(self._origin_l1doctests)
            # We insert the summarize infos into the header bar of the treeview
            self.insert_summarize_in_header_bar(summarize, self.header_bar)
            self.change_header_font() 
        
        self.update_row_height()
            
    def __add_verdicts_to_treeview(self, l1doctests:List[L1DocTest], parent=""):
        o_names = [c_l1doctest.get_name() for c_l1doctest in self._origin_l1doctests]
        
        for l1doctest in l1doctests:
            o_index = o_names.index(l1doctest.get_name())
            current_node = self._add_node_to_tree(self._origin_l1doctests[o_index], parent)
            if l1doctest.get_flag() == L1DocTestFlag.EMPTY_FLAG:
                self.treeview.insert(current_node, "end", values=l1doctest.get_node_lineno(), 
                                     text="No tests found", tags=("nonClickable", l1doctest.get_flag().get_color()))
            else:    
                self._add_verdicts_to_node(current_node, l1doctest.get_examples())
         
        self.__wrap_tree_content()
    
    def _add_node_to_tree(self, l1doctest: L1DocTest, parent=""):  
        flag: L1DocTestFlag = l1doctest.get_flag()
        open = get_option(OPEN_ONLY_RED_FUNCTIONS) and not get_option(CLOSE_FUNCTION_ROWS) \
                if get_option(OPEN_ONLY_RED_FUNCTIONS) and flag.is_failing() \
                else not get_option(OPEN_ONLY_RED_FUNCTIONS) and not get_option(CLOSE_FUNCTION_ROWS) 
        verdict_config = dict(text=self._get_l1doctest_stats(l1doctest),  
                              image=self.get_icon(flag.get_image_basename()), 
                              open=open)
        return self.treeview.insert(parent, "end", values=l1doctest.get_node_lineno(), tags=(CLICKABLE_TAG,), **verdict_config)
    
    def _add_verdicts_to_node(self, current_node:str, examples:list[Example]):
        """ 
        This function adds to the treeview all the rows that correspond 
        to the given ast node.
        """
        def __add_as_many_rows_as_text_lines(parent, text:str, lineno:int, tags:str):
            """
            Adds the necessary rows of each text to be inserted in the treeview.
            An exception message can be reported on several line, so we should 
            add as many rows as text lines. 
            """
            for line in text.splitlines():
                self.treeview.insert(parent, "end", text=line, values=lineno, tags=tags)

        verdicts_map = {
            FailedVerdict: (FAILED_RED_CHIP, False), # False means that the verdict is not an exception
            ExceptionVerdict: (EXCEPTION_RED_CHIP, True),
            PassedVerdict: (SUCCESS_GREEN_CHIP, False),
        }

        for example in examples:
            verdict = example.get_verdict()
            verdict_tags = (verdict.get_color(), CLICKABLE_TAG)
            item_text = str(verdict)

            for verdict_type, (icon, is_exception) in verdicts_map.items():
                if isinstance(verdict, verdict_type) and not isinstance(verdict, PassedSetupVerdict):                
                    if is_exception:
                        values = [verdict.get_lineno(), verdict.__class__.__name__, verdict.get_details()] 
                        verdict_tags += ("exception_as_link", "lightred" if get_option(HIGHLIGHT_EXCEPTIONS) else "")
                    else:
                        values = [verdict.get_lineno()]
                    # on insère le test dans la treeview   
                    current_test = self.insert_into_treeview(current_node, item_text, values, 
                                                        verdict_tags, icon, open=get_option(OPEN_RED_TEST_ROWS))
                    # on configure les tags pour insérer les détails de l'exception
                    if isinstance(verdict, FailedVerdict):
                        verdict_tags = (verdict_tags[0], "nonClickable")

                    if not is_exception:
                        __add_as_many_rows_as_text_lines(current_test, verdict.get_details(), verdict.get_lineno(), tags=verdict_tags)
                    break
    
    def insert_into_treeview(self, node="", text="", values=[], tags=(), image="", open=False):
        return self.treeview.insert(node, "end", text=text, values=values, 
                                    image=self.get_icon(image), tags=tags, open=open)
            
    def __check_if_editor_is_open(self) -> bool:
        """
            Returns True if an editor is already opened in thonny. 
            Otherwise, returns False 
        """
        return False if not self.workbench.get_editor_notebook().get_current_editor() else True
    
    def _get_l1doctest_stats(self, l1doctest: L1DocTest):
        """
        Get a string that represents how many tests are passed of the given l1doctest. 
        If the l1docstest is empty, returns only the name of the l1doctest.
        """
        # The first space is necessary so that all text will be aligned
        if l1doctest.get_flag() == L1DocTestFlag.EMPTY_FLAG:
            return " %s" % l1doctest.get_name()
        return " %s ~ %s/%s passed" %(l1doctest.get_name(), l1doctest.count_passed_tests(), l1doctest.count_tests())
        
    def insert_summarize_in_header_bar(self, summarize:Summarize, view: tktextext.TweakableText):
        """
        Builds the summarize test to be inserted in the header of the treeview.
        
        Args:
            summarize (Summarize): a named tuple that contains the summarize infos.
        """
        tests_run = "Tests run: %s\n" % summarize.total
        view.direct_insert("end", tests_run)
        
        insert_label = lambda label, color : view.direct_insert("end", label, tags=(color,)) 
        insert_how_many = lambda how_many, is_last=False: view.direct_insert("end", f"{how_many}" if is_last else f"{how_many}, ")
      
        insert_label("Success: ", "green")
        insert_how_many(summarize.success)
        
        insert_label("Failures: ", "darkred")
        insert_how_many(summarize.failures)
        
        insert_label("Errors: ", "darkred")
        insert_how_many(summarize.errors)
        
        insert_label("Empty: ", "orange")
        insert_how_many(summarize.empty, is_last=True)
            
    def build_summarize_object(self, l1doctests:List[L1DocTest]) -> Summarize:
        """
        Builds the summarize informations. 
        The summarize contains :
            - Total number of executed tests.
            - How many succeed tests, failed tests, error tests and empty tests.
        
        Args:
            results (List[L1DocTest]): all the l1doctests that have been evaluated.

        Returns:
            Summarize: a namedtuple that represents the summarize object.
        """       
        success = sum([l1doctest.count_passed_tests() for l1doctest in l1doctests])
        failures = sum([l1doctest.count_failed_tests() for l1doctest in l1doctests])
        errors = sum([l1doctest.count_error_tests() for l1doctest in l1doctests])
        empty = sum([1 for l1doctest in l1doctests if l1doctest.get_flag() == L1DocTestFlag.EMPTY_FLAG])
        total = success + failures + errors
        return Summarize(total, success, failures, errors, empty)
                   
    def clear_tree(self, clear_verdicts_data=False, clear_all=True, clear_errorview=False):
        """Clears the treeview by deleting all items. This method is called by
        the `update_tree_contents` method to clear the treeview before inserting
        the new rows.
        
        Note: this method is also called when the button `clear` is clicked. In 
        this case, the `event` is not None, then the original/copy lists of l1doctests
        will be cleared. Finally, the treeview or/and header or/and the errorview 
        will be cleared.
        
        Args:
            clear_verdicts_data: If true, the original and copy lists of l1doctests will be cleared.
            This argument is set to True when the button `clear` is clicked.
            clear_all: If True, the treeview and its header will be cleaned.
            clear_errorview: if True, the error view will be cleaned.            
        """
        if clear_verdicts_data: # si l'event est déclenché par le bouton `clear` alors on vide les listes
            from thonnycontrib.l1test_frontend import get_l1test_runner
            self._origin_l1doctests.clear()
            self._copy_l1doctests.clear()
            clear_errorview = True
            get_l1test_runner().set_has_exception(False)
        if clear_all:
            self.clear_header_bar()  # on supprime le contenu du header
        if clear_errorview:
            error_view:L1TestErrorView = self.workbench.get_view(L1TestErrorView.__name__)
            error_view.clear()
        self.treeview.delete(*self.treeview.get_children())
        self.__init_special_attributes()
        self.disable_menu()
        
    def clear_header_bar(self):
        """Clears the header of the treeview."""
        if self.header_bar:
            self.header_bar.direct_delete("1.0", "end")
            self.resize_header_bar()
    
    def _remove_highlight_selection_effect(self, event=None):
        """
        This function remove the selection effect. When a treeview's row is selected
        it removes the highlight effect on the selected row. So the selected row 
        will look like it is not selected.
        """
        self.style.map('Treeview', background=[], foreground=[])
    
    def _highlight_line_on_editor(self, lineno:int, editor: CodeView):
        """Highlights the line in the editor that corresponds to the selected row in the treeview.

        Args:
            lineno (int): the line number to highlight.
            editor (CodeView): the editor where the line will be highlighted.
        """
        index = editor.text.index(str(lineno) + ".0")
        editor.text.see(index)  # make sure that the double-clicked item is visible
        editor.text.select_lines(lineno, lineno)
    
    def _restore_row_selection_effect(self):
        """
        This function show the selection effect. When a treeview's row is selected
        it shows the highlight effect on the selected row. 
        """
        self.style.map('Treeview', 
                       background=[('selected', 'focus', '#ADD8E6'), ('selected', '!focus', '#D3D3D3')], 
                       foreground=[('selected', 'focus', 'black'), ('selected', '!focus', 'black')])

    
    def _on_select(self, event=None):
        """
        When a row is selected this function will be triggered. This function highlights 
        the line in the editor that corresponds to the seelcted row in the treeview. 
        """
        self._restore_row_selection_effect()
        editor = self.workbench.get_editor_notebook().get_current_editor()
        if editor:
            code_view = editor.get_code_view()
            focus = self.treeview.focus()
            if not focus: 
                return

            item = self.treeview.item(focus)
            values = item["values"]
            if not values:
                return
                
            if ExceptionVerdict.__name__ in values:
                # side_by_side=True means that the error view will be displayed side by side with the treeview
                self.workbench.event_generate(L1TREE_VIEW_EVENT, error_title=item["text"],
                                                error_msg=values[-1], side_by_side=True) 
                self.treeview.focus_set()
                        
            lineno = values[0]
            self._highlight_line_on_editor(lineno, code_view)
            self.workbench.event_generate(
                "OutlineDoubleClick", item_text=self.treeview.item(self.treeview.focus(), option="text")
            )

    def disable_menu(self):
        """disable the menu button of the treeview"""
        self.menu_button.state([tk.DISABLED])
        
    def enable_menu(self):
        """enable the menu button of the treeview"""
        self.menu_button.state(["!disabled"])
     
    def insert_in_header(self, text, image:str|tk.PhotoImage=None, clear=False, tags=tuple()):
        """ 
        Inserts text in the header of the treeview. 
        Args:
            text: the text to insert
            image: the basename with it's extension of an image to insert. 
            For example: "info.png". The image must be in the folder `docs/res`.
            clear: if True, the header will be cleared before inserting the text
            tags: the tags to apply to the text. For example: ("red",)
        """
        if clear:
            self.header_bar.direct_delete("1.0", tk.END)
        if image:
            if isinstance(image, str):
                image = self.get_icon(image)
            self.header_bar.image_create(tk.END, image=image)
            text = " " + text # add a space after the image
        self.header_bar.direct_insert(tk.END, text, tags=tags)
        self.resize_header_bar()
           
    def is_empty(self):
        return len(self.treeview.get_children()) == 0
    
    def is_header_bar_cleared(self): 
        return not self.header_bar.get("1.0", tk.END).strip("\n")    
    
    def hide_view(self):
        self.workbench.hide_view(self.__class__.__name__)
    
    def show_view(self):
        self.workbench.show_view(self.__class__.__name__)
        
    def set_verdicts(self, verdicts: List[L1DocTest]):
        self._origin_l1doctests = verdicts
        self._copy_l1doctests = self._origin_l1doctests.copy()
    
    def get_original_verdicts(self):
        """ Returns the original list of all l1doctests. """
        return self._origin_l1doctests
    
    def get_copy_verdicts(self):
        """ Returns the copy list of all l1doctests. """
        return self._copy_l1doctests
    
    def get_treeview(self):
        return self.treeview
    
    def get_icon(self, ref_name:str):
        """
        Returns the image reference saved in the treeview. Returns None if the 
        image reference is not found.
        """
        if ref_name in self.image_references:
            return self.image_references[ref_name]
        return None
    
    def get_max_lines(self):
        """Returns the maximum number of lines of the longest wrapped line of the treeview."""
        return self.__max_lines



from types import ModuleType
import unittest as ut
from thonnycontrib.tests.fixtures.backend_mock import *
from thonnycontrib.backend.doctest_parser import ExampleWithoutExpected
from unittest.mock import *
from thonnycontrib.backend.evaluator import Evaluator
from thonnycontrib.exceptions import *
from thonnycontrib.backend.verdicts.PassedSetupVerdict import PassedSetupVerdict
from thonnycontrib.backend.ast_parser import *
from thonnycontrib.backend.verdicts.ExceptionVerdict import ExceptionVerdict
from thonnycontrib.backend.doctest_parser import *

# #######################################################
#    Tous les tests qui suivent utilisent l'invite `$$$`
#               SANS valeur attendue  
# #######################################################
class TestEvaluator(ut.TestCase):
    def setUp(self):
        self.evaluator = Evaluator(filename="<string>")
        self.mock_backend = backend_patch.start()
    
    def tearDown(self) -> None:
        del self.evaluator
        backend_patch.stop()
    
    """
    Ce test vérifie:
    1. Le type `ExampleWithoutExpected` est le type `Example` extrait 
    par le doctest parser.
    2. S'il existe que des setups réussi dans une fonction alors le l1doctest associé aura 
    le flag EMPTY_FLAG. On assure aussi que le verdict associé à l'example est PassedSetupVerdict
    """
    def test_evaluate_when_setup(self):
        fake_source = \
"""
def f(a, b):
    '''
    $$$ a = 0
    '''
    if a < 0 and b < 0:
        return None
    return a + b
""" 
        
        self.evaluator.set_source(fake_source)
        self.evaluator.set_module(self.__build_module_from_source(fake_source))
        
        # ###########################################################
        # ------------- Vérification du type Example ----------------
        self._assert_example_type(ExampleWithoutExpected) # assure qu'il y a un seul example
        
        # ###################################################
        # ------------- Vérification du verdict -------------
        l1doctests = self.evaluator.evaluate()
    
        # on assure qu'il existe un seul l1doctest avec ses examples
        self.assertTrue(len(l1doctests) == 1)
        l1doctest = l1doctests[0]
        
        # on assure que l1doctest ne contient aucun tests
        self.assertTrue(l1doctest.has_only_setUps())
        self.assertTrue(l1doctest.is_evaluated()) # s'assure que le l1doctest a été évalué
        self.assertEqual(l1doctest.get_flag(), L1DocTestFlag.EMPTY_FLAG)
        
        # on assure que le verdict de l'example est un setUp qui réussit
        examples = l1doctest.get_examples()
        self.assertTrue(len(examples) == 1) # on assure qu'il existe un seul type Example
        self.assertTrue(isinstance(l1doctest.get_verdict_from_example(examples[0]), PassedSetupVerdict))   
    
    """
    Ce test vérifie:
    1. Le type `ExampleWithoutExpected` est le type `Example` extrait 
    par le doctest parser.
    2. S'il existe une erreur de syntaxe au niveau du setup alors
    un verdict `ExceptionTest` sera renvoyé. Le flag associé au l1doctest sera alors FAILED_FLAG
    """
    def test_evaluate_when_syntax_error_on_setup(self):  
        fake_source = \
"""
def f(a, b):
    '''
    $$$ a = 
    '''
    if a < 0 and b < 0:
        return None
    return a + b
""" 
        
        self.evaluator.set_source(fake_source)
        self.evaluator.set_module(self.__build_module_from_source(fake_source))
        
        # ###########################################################
        # ------------- Vérification du type Example ----------------
        self._assert_example_type(ExampleWithoutExpected)
        
        # ###################################################
        # ------------- Vérification du verdict -------------
        l1doctests = self.evaluator.evaluate()

        # on assure qu'il existe un seul l1doctest avec ses examples
        self.assertTrue(len(l1doctests) == 1)
        l1doctest = l1doctests[0]
        
        # on assure que l1doctest contient aucun test, mais qu'il exisite une erreur
        self.assertTrue(l1doctest.has_only_setUps())
        self.assertTrue(l1doctest.is_evaluated()) # s'assure que le l1doctest a été évalué
        self.assertEqual(l1doctest.get_flag(), L1DocTestFlag.FAILED_FLAG)
        
        # on assure que le verdict de l'example est en erreur
        examples = l1doctest.get_examples()
        self.assertTrue(len(examples) == 1) # on assure qu'il existe un seul type Example
        
        verdict = l1doctest.get_verdict_from_example(examples[0])
        self.assertTrue(isinstance(verdict, ExceptionVerdict))   
        self.assertTrue(SyntaxError.__name__ in verdict.get_details()) 
    
    def _assert_example_type(self, example_type: Example):
        """Assert that the expected `Example` type is of type the given `example_type`"""
        l1doctests = self.evaluator.get_test_finder().find_l1doctests()
        
        # on assure qu'il existe un seul l1doctest avec ses examples
        self.assertTrue(len(l1doctests) == 1)
        L1docTest = l1doctests[0]
        examples = L1docTest.get_examples()
        self.assertTrue(len(examples) == 1) # on assure qu'il existe un seul type Example 
        self.assertTrue(isinstance(examples[0], example_type))
    
    def __build_module_from_source(self, source: str) -> ModuleType:
        """
        Build a module containing the functions declared in the given `source`.
        """
        from types import ModuleType
        fake_module = ModuleType(self.evaluator.get_filename())
        exec(source, fake_module.__dict__)
        return fake_module


if __name__ == '__main__':
    ut.main(verbosity=2)   
        


PLUGIN_NAME = "L1Test"
DEBUG_NAME = "L1Test debugger"
ERROR_VIEW_LABEL = '%s errors' % PLUGIN_NAME

# ici vous pouvez changer la syntaxe du doctest. 
# version 2022 PJI : Actuellement on garde la syntaxe `$py`.
L1TEST_SYMBOL = "[$]py"
# version 2022 avant la rentrée
L1TEST_SYMBOL1 = "[$][$][$]"
L1TEST_SYMBOL2 = "[$]PY"
L1TEST_SYMBOL3 = "[$]py"
# L'invite des tests qui vérifient la levée d'exception
L1TEST_EXCEPTION_SYMBOL = "[$][$]e"


# ############################################################################################### #
#                       LES NOUVELLES VARIABLES VERSION 2023 PFE                                  #
# ############################################################################################### #

# Le nom de la commande magique pour l1test(doit toujours commencer par une majuscule)
BACKEND_COMMAND = "L1test"


# ############ Les noms des clés du dictionnaire renvoyé par le l1test_backend ############
# Le nom de l'attribut contenant les résulats des tests renvoyés par l1test_backend
VERDICTS = "verdicts"
# Le nom de l'attribut contenant une exception levée et renvoyée par l1test_backend
L1TEST_EXCEPTION = "l1test_exception"


# ############ Les labels des buttons du menu l1test treeview ############
PLACE_RED_TEST_ON_TOP_LABEL = "Place the red tests on the top"
RESUME_ORIGINAL_ORDER = "Resume original order"
SHOW_ONLY_RED_TESTS = "Show only red tests"
SHOW_ALL_TESTS = "Show all the tests"
GROUP_BY_VERDICTS = "Group by verdicts"
EXPAND_ALL = "Expand all"
FOLD_ALL = "Fold all"
UPDATE_FONT_LABEL = "Update the font"
INCREASE_SPACE_BETWEEN_ROWS = "Inrease row height"
DECREASE_SPACE_BETWEEN_ROWS = "Decrease row height"
CLEAR_LABEL = "Clear"

# Le message affiché sur la treeview quand `l1test` est en cours d'execution
L1TEST_IN_PROGRESS = "Executing tests in progress"

# the evaluation states
PENDING_STATE = "Pending" 
EXECUTED_STATE = "Executed" 
FINISHED_STATE = "Finished"

# The title of the error view when the docstring genertor shows the raised error
CANNOT_GENERATE_THE_DOCSTRING = "Cannot generate the docstring :"
# The title of the error view when the l1test shows the raised error
CANNOT_RUN_TESTS_MSG = "Cannot run %s :" %(PLUGIN_NAME)

# A special event that `L1TestTreeview` sends to `L1TestRunner` when clicking on an exception test
# The event transfers the details of the clicked exception to the `L1TestRunner` 
# which will show it in the error view.
L1TREE_VIEW_EVENT = "L1TreeviewEvent"

# Les images utilisées par la treeview
PENDING_ICON = "pending_icon.png"
ERROR_ICON = "error_icon.png"
RESTART_ICON = "restart_icon.png"
FAILED_RED_CHIP = "failed_red_chip.png" # le petit cercle rouge qui précède un test qui a échoué
EXCEPTION_RED_CHIP = "exception_red_chip.png" # le petit cercle rouge (avec un poitn d'exclamation) qui précède un test qui a échoué
SUCCESS_GREEN_CHIP = "success_green_chip.png" # le petit cercle vert qui précède un test qui a réussi


from abc import *
import ast
from collections import namedtuple

# Un objet pour représenter le vocabulaire utilisé par le template 
class TemplateVocabulary:
    def __init__(self, todo_label="", summary_label="", param_label="",
                cu_label="", rtype_label="", rvalue_label="", test_label="", **kw) -> None:
        self.todo_label = todo_label
        self.summary_label = summary_label
        self.param_label = param_label,
        self.cu_label = cu_label,
        self.rtype_label = rtype_label,
        self.rvalue_label = rvalue_label,
        self.test_label = test_label
        
        # Le vocabulaire est flexible pour introduire d'autre mots
        for attribute, value in kw.items():
            setattr(self, attribute, value)
        
class DocTemplate(ABC):
    # Ces constantes peuvent être utilisées dans les classes d'implémention
    NEW_LINE = "\n"
    DOCSTRING_SYMBOL = '"""'
    
    def __init__(self, vocabulary=TemplateVocabulary()) -> None:
        vocabulary.todo_label = ""
        vocabulary.summary_label = "x_résumé_x"
        vocabulary.param_label = "Paramètres :"
        vocabulary.test_label = "Exemples :\n$$$ "
        vocabulary.cu_label = "Contraintes d'utilisation : "
        self.vocabulary = vocabulary
                            
    @abstractmethod
    def _format_general_summary(self) -> str:
        """
        Returns:
            str: Returns a label which will indicate to write a summary of the function.
        """
        pass
    
    @abstractmethod
    def _format_params(self, params) -> str:
        """
        Args:
            params (List): It's a list of the arguments.

        Returns:
            str: Returns the parameter representation section of a node in a docstring. 
        """
        pass
    
    @abstractmethod
    def _format_usage_constraints(self) -> str:
        """
        Returns:
            str: Returns the usage constraints representation section in a docstring.
        """
        pass
    
    @abstractmethod
    def _format_return_value(self) -> str:
        """
        Returns:
            str: Returns the return value representation section in a docstring.
        """
        pass
    
    @abstractmethod
    def _format_test_examples(self) -> str:
        """
        Returns:
            str: Returns the test examples representation section in a docstring.
        """
        pass
    
    @abstractmethod
    def get_template(self, node:ast.AST=None) -> str:
        """Build the complete docstring template. 
        This method must invoke the above abstract methods.
        
        Args:
            node (ast.AST): The AST node in which the dosctring will be generated.

        Returns:
            str: Returns the template representation. 
        """
        pass  
    
    @abstractmethod
    def get_id_signature(self) -> str: 
        pass      
    
class DocFunctionTemplate(DocTemplate):
    '''
    Modifié pour coller au cours d'Info L1
    '''
    def __init__(self) -> None:
        super().__init__()
        self.vocabulary.rvalue_label = "Valeur de retour "
        self.vocabulary.rtype_label = "(%s) :"
        
    def get_parameters(self, node:ast.AST):
        """
        Get the paramters of a given node.
        
        Args:
            node (ast.AST): An AST node. Must be an ast.FunctionDef or ast.AsyncFunctionDef

        Returns:
            List: Returns a List of arguments of the given node.
        """
        if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            return node.args.args
        return []

    def _format_general_summary(self):
        return self.vocabulary.summary_label + self.NEW_LINE
    
    def _format_params(self, params):
        if params is None:
            return ""
        args_to_exclude = ["self", "cls"]
        label = self.vocabulary.param_label + self.NEW_LINE
        format_params = ""
        for p in params:
            arg_type = ast.unparse(p.annotation) if p.annotation else ""     
            arg_name = p.arg 
            if arg_name not in args_to_exclude: 
                format_params += "- %s (%s) : %s\n" %(arg_name, arg_type, self.vocabulary.todo_label)
        return label + format_params
    
    def _format_usage_constraints(self):
        return self.vocabulary.cu_label + self.vocabulary.todo_label + self.NEW_LINE   

    def _format_return_value(self, node: ast):
        return_type = ast.unparse(node.returns) if node.returns else ""
        return_descr = self.vocabulary.rtype_label % return_type + self.NEW_LINE
        return self.vocabulary.rvalue_label  + return_descr
    
    def _format_test_examples(self):
        label = self.vocabulary.test_label + self.NEW_LINE
        todo = self.vocabulary.todo_label + self.NEW_LINE
        return label + todo
        
    def get_template(self, node: ast.AST):
        return (
            self.DOCSTRING_SYMBOL + 
            self._format_general_summary() + self.NEW_LINE + 
            self._format_params(self.get_parameters(node))  + 
            self._format_return_value(node) + 
            self._format_usage_constraints() +
            self._format_test_examples() + 
            self.DOCSTRING_SYMBOL + self.NEW_LINE
        )
    
    def get_id_signature(self): 
        return "def" 

class DocClassTemplate(DocTemplate): 
    def _format_general_summary(self):
        return self.vocabulary.summary_label + self.NEW_LINE
    
    def _format_params(self):
        return self.vocabulary.param_label
     
    def _format_usage_constraints(self):
        return (self.vocabulary.cu_label +
                self.vocabulary.todo_label +
                self.NEW_LINE 
            )  

    def _format_return_value(self):
        return self.vocabulary.rvalue_label
    
    def _format_test_examples(self):
        label = self.vocabulary.test_label + self.NEW_LINE
        todo = self.vocabulary.todo_label + self.NEW_LINE
        return label + todo
            
    def get_template(self, node):
        return self.DOCSTRING_SYMBOL + \
               self._format_general_summary() + self.NEW_LINE + \
               self._format_usage_constraints() + \
               self._format_test_examples() + \
               self.DOCSTRING_SYMBOL + self.NEW_LINE

    def get_id_signature(self): 
        return "class" 

class DocTemplateFactory:            
    @staticmethod
    def create_template(type:str):
        return DocTemplateFactory.__search_type(type)
    
    @staticmethod
    def __docTemplate_subclasses(cls=DocTemplate):
        return set(cls.__subclasses__()) \
               .union([s for c in cls.__subclasses__() \
                            for s in DocTemplateFactory.__docTemplate_subclasses(c)])
    
    @staticmethod
    def __search_type(type:str) -> DocTemplate|None:
        template_types = DocTemplateFactory.__docTemplate_subclasses()
        
        find_type = [t() for t in template_types if type==t().get_id_signature()]
        return find_type[0] if find_type else None   



from typing import List
from .l1test_error_view import L1TestErrorView
from .l1test_treeview import L1TestTreeView
from ..properties import *
from ..l1test_configuration.l1test_options import *
from ..backend.ast_parser import L1DocTest
from thonny.codeview import *
from ..utils import format_filename_to_hyperlink, remove_url_part
import thonny

class L1TestReporter():    
    def __init__(self, main_view_class=L1TestTreeView, error_view_class=L1TestErrorView):
        self.__treeview_class = main_view_class
        self.__error_view_class = error_view_class 
        
        self._workbench = thonny.get_workbench()

        self.__treeview: L1TestTreeView = self._workbench.get_view(self.__treeview_class.__name__)
        self.__error_view: L1TestErrorView = self._workbench.get_view(self.__error_view_class.__name__)
    
    def display_error_msg(self, error_msg:str, title=CANNOT_RUN_TESTS_MSG, color="red"):
        """Display an error message into the `L1TestErrorView`.
        The `prefix_info` is the title preceding the error message.
        
        Args:
            error_msg (str): the error message to be displayed
            prefix_info (str, optional): the title preceding the error message. 
            Defaults to CANNOT_RUN_TESTS_MSG.
            color (str, optional): the color of the whole content. Defaults to "red".
        """
        self.__error_view.clear()
        if title:
            title = title if title.endswith("\n") else title + "\n"
            self.__error_view.append_text(title, tags=("darkred", "title"))

        hyperlink_error = format_filename_to_hyperlink(error_msg)

        if hyperlink_error: 
            self.__error_view.append_rst(hyperlink_error) # hyperlink doit être ajouté en mode RST
            error_msg = remove_url_part(error_msg) 
            self.__error_view.append_text(error_msg, tags=color) # le reste du message est ajouté en mode texte
        else:
            self.__error_view.append_text(error_msg, tags=color)
                       
    def display_tests_results(self, verdicts: List[L1DocTest]):
        """Display the verdicts into the `L1TestTreeView`.

        Args:
            verdicts (List[L1DocTest]): the dictionary of verdicts.
        """
        self.__treeview.set_verdicts(verdicts)
        self.__treeview.update_tree_contents(verdicts)
    
    def set_treeview_class(self, treeview_class):
        self.__treeview_class = treeview_class   
    
    def get_error_view_class(self, error_view_class):
        self.__error_view_class = error_view_class      
    
    def get_treeview(self) -> L1TestTreeView:
        return self.__treeview    
    
    def get_error_view(self) -> L1TestErrorView:
        return self.__error_view    
    


import unittest as ut
from thonnycontrib.docstring_generator.doc_generator import *

class MockDocGeneratorStrategy(DocGenerationStrategy):
    def __init__(self):
        super().__init__(text_widget=None)
        
    def can_generate(self) -> bool:
        pass
    
    def generate(self) -> None:
        pass
    
class TestDocGenerator(ut.TestCase):
    def setUp(self):
        self.docGenerationStrategy = MockDocGeneratorStrategy()
    
    def tearDown(self) -> None:
        del self.docGenerationStrategy
    
    def test_doc_generator_when_ok(self):
        signature = "def func():"
        generated = self.docGenerationStrategy._generate(signature)
        
        self.assertTrue(len(generated) > 0)
        
        expected_doc = \
'''\
    """x_résumé_x

    Paramètres :
    Valeur de retour () :
    Contraintes d'utilisation : 
    Exemples :
    $$$ 

    """
'''
        self.assertTrue(generated == expected_doc)
    
    def test_doc_generator_typing_support_case_1(self):
        arg_type, r_type = "str", "int"
        signature = "def func(a: %s) -> %s:" % (arg_type, r_type)
        generated = self.docGenerationStrategy._generate(signature)
        
        self.assertTrue(len(generated) > 0)
        self.assertTrue(arg_type in generated)
        self.assertTrue(r_type in generated)
    
    def test_doc_generator_typing_support_case_2(self):
        arg_type = "str"
        signature = "def func(a: %s):" % arg_type
        generated = self.docGenerationStrategy._generate(signature)
        
        self.assertTrue(len(generated) > 0)
        self.assertTrue(arg_type in generated)
    
    def test_doc_generator_when_syntax_error(self):
        signature = "def func):"
       
        with self.assertRaises(DocGeneratorParserException) as e:
            self.docGenerationStrategy._generate(signature)
        
        raised_exception = e.exception      
        self.assertTrue(SyntaxError.__name__ in str(raised_exception))
        
    def test_doc_generator_when_signature_doesnt_finish_with_colon(self):
        signature = "def func()"
        
        with self.assertRaises(DocGeneratorParserException) as e:
            self.docGenerationStrategy._generate(signature)
        
        raised_exception = e.exception      
        self.assertTrue(SyntaxError.__name__ in str(raised_exception))
        self.assertTrue(SyntaxError.__name__+": expected ':'" in str(raised_exception))
        
if __name__ == '__main__':
    ut.main(verbosity=2)   


'''
Permet de générer pour le plugin de log une synthèse des tests exécutés.
Corentin.
'''

from typing import List
from thonny import get_workbench
from .backend.ast_parser import L1DocTest

#Attention si le nom change ici, il faut aussi le changer dans Thonny-LoggingPlugin
NOM_EVENT_TEST = "l1Tests"
NOM_EVENT_DOC = "l1Tests.DocGenerator"
wb = get_workbench() 

# Modifié par Mathieu Bachelet, PJI 2022-2023, mais sur la V2 de L1test
# Modifié par Mirabelle pour adaptation à la V3 de L1test
def log_in_thonny(test_results: List[L1DocTest], filename:str, selected):
    tests_data = []
    for l1doctest in test_results:
        for example in l1doctest.get_examples():
            # test:Example
            verdict = example.get_verdict() # type ExampleVerdict
            res = vars(verdict).copy() # pour récupérer les attributs filename, lineno, tested_line, expected_results, details
            res['details'] = str(res['details']) # sinon on peut récupérer un entier
            res['verdict'] = type(verdict).__name__ # pour connaître le verdict, qui n'est pas dans une champ quelconque mais ds le type lui-même
            res['name'] = l1doctest.get_name() # nom fonction ou classe
            res['status'] = verdict.isSuccess() # True si test au vert
            tests_data.append(res)
    wb.event_generate(NOM_EVENT_TEST, None, selected=selected, tests=tests_data, filename=filename)


def log_doc_in_thonny(node):
    if wb:
        wb.event_generate(NOM_EVENT_DOC,None, name = node.name)


#La fonction anonyme car il faut une fonction pour bind, avec un argument parce qu'elle reçoit l'événement. 
if wb:   
    wb.bind(NOM_EVENT_TEST, lambda x : 0, True)
    wb.bind(NOM_EVENT_DOC, lambda x : 0, True)



from io import StringIO
from types import ModuleType
from thonny import get_workbench
from thonnycontrib.exceptions import CannotImportModuleException, CompilationError
from .properties import BACKEND_COMMAND 
from .environement_vars import *
import os, re, ast, traceback, textwrap, tkinter as tk
import thonny
from thonny.editors import EditorCodeViewText

def wrap(string:str, length=8, break_long_words=False):   
    """Wrap a text using the `wraptext` module.

    Args:
        string (str): the string to wrap.
        length (int, optional): the min length from which the string will be wrapped. Defaults to 8.
        break_long_words (bool, optional): if False the entire words will not be wrapped
                                        if the words lentgh is more that the given length. Defaults to False.
        separator (str, optional): the separator around which to wrap the filenames.
    Returns:
        str: the wrapped string.
    """
    return '\n'.join(textwrap.wrap(string, length, break_on_hyphens=False, break_long_words=break_long_words))


def create_node_representation(node:ast.AST):
    """
    Returns the node representation. Especially, returns the prototype or the signature of the node.
    This function can only construct a string representation of the supported nodes. 
    The supported nodes are reported in ASTParser.py in the global variable SUPPORTED_TYPES.
    
    Even if unsupported node is given so just it's name is returned.
    
    Args: 
        node (ast.AST): The supported node 

    Returns:
        str: Return the string represantation of a node
    """
    arg_to_exclude = lambda arg: arg in ("self", "cls")
    if isinstance(node, ast.ClassDef):
        return "%s(%s)" % (node.name, ", ".join([base.id for base in node.bases]))
    elif isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
        return "%s(%s)" % (node.name, ", ".join([a.arg for a in node.args.args if not arg_to_exclude(a.arg)]))
    else:
        return node.name

def remove_url_part(error_msg:str):
    """Removes the hyperlink section from the `error_msg`. The hyperlink
    contains the filename and line number. 
    
    This function is only used by the L1TestErrorView to show the details of
    an error as a normal text.

    Args:
        error_msg (str): the error message from which the hyperlink will
                        be removed.

    Returns:
        str: The new error message without hyperlink part.
    """
    return re.sub(r'(File .*,\s*line\s*\d*)(, .*)?', "", error_msg)

def replace_error_line(error:str, line_number:int):
    """Replace the line of the error to iven `line_number`

    Args:
        error (str): The error that probably contains the error line.
        line_number (int): the new line number of the error.

    Returns:
        str: the same error with a new error line number.
    """
    return re.sub(r'(?P<match>line) (\d+)', '\g<match> %s' % line_number, error)

def replace_filename(filename, error:str):
    """Replace the filename in the given error to the specified `filename`

    Args:
        filename (str): the new filename
        error (str): the error message

    Returns:
        str: the same error but with replaced filename.
    """
    return re.sub(r'"(<.*>)?"', '"%s"' % filename, error)

def format_filename_to_hyperlink(error_message:str):
    """
    Finds the filename and line number in the given error message 
    then creates the corresponding hyperlink leading to the right place in the code editor.
    
    If no filename and line number found so an empty string is returned
    
    Args:
        text (str): the error_message supposed containing the hyperlink.

    Returns:
        str: Returns a hyperlink in RST format. 
    """
    def extract_file_and_lineno(error):
        file_line = re.search(r'File\s*(.*),\s*line\s*(\d*)(.*)', error)
        return {
                "filename": file_line.group(1).replace("\"", ""), 
                "lineno": file_line.group(2),
                "complement": file_line.group(3)
            } if file_line else None
    
    extracted = extract_file_and_lineno(error_message)
    if extracted:
        if os.path.exists(extracted["filename"]):
            filename, lineno, comp = extracted["filename"], extracted["lineno"], extracted["complement"]
            url = "thonny-editor://" + escape(filename).replace(" ", "%20")
            if lineno is not None:
                url += "#" + str(lineno)
            # return the hyperlink format following the rst specification
            return "`File \"%s\", line %s%s <%s>`__\n" % (escape(filename), lineno, comp, url)
        else: 
            return ""
    else:
        return ""

def escape(s:str):
    return (
        s.replace("\\", "\\\\")
         .replace("*", "\\*")
         .replace("`", "\\`")
         .replace("_", "\\_")
         .replace("..", "\\..")
    )

import traceback
import re
from io import StringIO

def get_last_exception(exc_info:tuple):
    """
    Formats exception information provided by ~exc_info~.
    Extracts the last exception located in the last frame.
    
    Args:
        exc_info (tuple): The tuple must contain three elements: 
                        (type, value, traceback) as returned by sys.exc_info().
        with_details (bool): Set as True and some extra informations will be considered
                        while formatting the exception informations. Defaults to True.

    Returns:
        str: Returns a string containing a traceback message for the given ~exc_info~.
    """
    # Get a traceback message.
    excout = StringIO()
    exc_type, exc_val, exc_tb = exc_info
    traceback.print_exception(exc_type, exc_val, exc_tb, file=excout)
    # la variable `content`` va contenir toute la trace liée à l'exception levée
    content = excout.getvalue()
    excout.close()

    # mais on en retire que la dernière frame qui indiquent l'exception renvoyée par l'éditeur thonny
    last_exception = __extract_last_exception(content.rstrip("\n"))
    
    # parfois Python inclut plusieurs lignes vides dans l'exception.
    last_exception = re.sub("\n+", "\n", last_exception)
    # On assure que tout ce qui procède la dernière exception est retiré
    return re.sub(r'.*(?=File.*)', "", last_exception)

def __extract_last_exception(traceback_content:str):
    """Extract the last exception from the traceback frames,

    Args:
        content (str): The returned traceback as string.

    Returns:
        str: The lines representing the compilation error 
    """
    # splitted va contenir toutes les ligne de la traceback
    splitted = traceback_content.split("\n")
    keyword = "File"
    last_frame_index = 0
    
    # On veut pas afficher les frames du backend sur thonny, c'est inutile et illisible.
    frames_to_exclude = ["cpython_backend", os.path.join("thonny", "backend")]
    
    # on filtre par le mot "File" pour récupérer la dernière Frame.
    # C'est la dernière frame qui contient l'erreur survenu du script de thonny 
    # On veut pas, évidement, afficher toute la trace contenant les exceptions levées par l1test backend.
    for i in range(len(splitted)):
        frame = splitted[i]
        if keyword in frame :    
            last_frame_index = i
            
    # last_frame contient le nom du fichier et le détail de l'erreur              
    last_frame = splitted[last_frame_index:] 
    # Le nom du fichier de la dernière frame est à la position 0
    last_frame_file = last_frame[0]
    
    # Dans le cas de l'interuption du programme par un <Control+c>, 
    # la dernière frame contiendra une exception levée par le backend thonny.
    # En l'occurrence, on doit pas afficher toute la frame, mais juste le message renvoyé.
    # En général, on évite d'afficher les frames survenue du backend thonny.
    for frame_to_exclude in frames_to_exclude: 
        if frame_to_exclude in last_frame_file:
            last_frame_index = -1    
            
    # last_frame[-1] -> contient le message de l'exception de la frame          
    return last_frame[-1] if last_frame_index < 0 else "\n".join(last_frame).strip()

def get_module_name(filename:str):
    """
    Gets the module name from the specified filename.
    This function simply gets the basename of the filename, then removes 
    the ".py" extension. 

    Args:
        filename (str): an absolute path

    Returns:
        str: the module name of the given filename.
    """
    return  get_basename(filename, "py")

def get_basename(filename:str, extension:str="*"):
    """
    Gets the module name from the specified filename.
    This function simply gets the basename of the filename, then removes 
    the given extension. 

    Args:
        filename (str): an absolute path
        extension (str): the extension to remove from the filename. 
        Extension can be a regex. Use '*' to remove all extensions.

    Returns:
        str: the basename of the given filename without the given extension.
    """
    return re.sub(r'.'+extension, "", os.path.basename(filename))
   
def import_module(filename:str) -> ModuleType:
    """
    Import a module from a given filename. 
    The ~filename~ can be also the absolute path of a file.
    
    This function can raise an exception if the imported module 
    contains a compilation error. You should catch it somewhere.
    
    Args:
        filename (str): The filename(or the absolute path) of a file

    Returns:
        ModuleType: Returns the corresponding module of the given file.
        
    Raises: 
        CannotImportModuleException: if the filename cannot be imported by the 
        importlib module.
        CompilationError: any other exception related to a compilation error.
    """
    import importlib.util as iu, os, sys
    
    # import the module specification. 
    # To learn more about ModuleSpec `https://peps.python.org/pep-0451/`    
    module_name = get_module_name(filename)
    spec = iu.spec_from_file_location(module_name, filename)
    if not spec: 
        if not filename.endswith(".py"):
            msg_error = "The file \"%s\" cannot be imported.\n\n" % filename + \
                        "Please, be sure that the extension of the file is `.py`" 
        else:
            msg_error = "Cannot found the file `%s`" %filename 
        msg_error = "Error when importing the module \"%s\":\n" % module_name + msg_error
        raise CannotImportModuleException(msg_error)
    
    imported_source = iu.module_from_spec(spec)
    
    workingdir = os.path.split(imported_source.__file__)
    if (len(workingdir) > 0):
        basedir = workingdir[0]
        dirs = get_all_parent_directories(basedir)
        
        # ajout des packages parents au sys.path
        # on fait ça parce que on veut assurer les imports 
        # des fichiers contenant dans les packages parents
        [sys.path.append(path) for path in dirs]
        
        # ajout des sous packages au sys.path 
        # pour assurer les imports des fichiers contenant dans les sous packages
        sub_packages = get_sub_directories(basedir)
        [sys.path.append(basedir + os.sep + path) for path in sub_packages]
  
    try:
        # This line can raise an exception if the module contains compilation errors,
        # or if the module imports non existed modules
        spec.loader.exec_module(imported_source) 
        return imported_source
    except BaseException as e:
        # the compilation error is catched and raised as a CompilationError
        # and the evaluation is interrupted(because we cannot parse a content
        # with compilation errors).
        error_info = sys.exc_info()
        formatted_error = get_last_exception(error_info)
        raise CompilationError(formatted_error) 

def get_all_parent_directories(dir_path:str):
    """
    For a given path of a directory returns all the parents directory from that path.
    
    Examples:
    >>> get_all_parent_directories('/home/stuff/src')
    ['/home', '/home/stuff', '/home/stuff/src']
    """
    import os
    
    if dir_path is None:
        return []
    
    dirs = dir_path.split(os.sep)
    m = ""
    res = []
    for e in dirs[1:]:
        m += os.sep + e  
        res.append(m)
    return res

def get_sub_directories(package_path:str):
    """
        Get all sub-packages of a given package(param). 
    """
    from setuptools import find_packages
    return [p.replace(".", os.sep) for p in find_packages(package_path)]

def get_focused_writable_text():
    """
    Returns the focused text

    Returns:
        Widget: A widget Object if there's a focused area in the editor
        None : if no focused area exists in the editor
    """
    from thonny.editors import EditorCodeViewText
    widget = get_workbench().focus_get()
    # In Ubuntu when moving from one menu to another, this may give None when text is actually focused
    if isinstance(widget, EditorCodeViewText) and (
        not hasattr(widget, "is_read_only") or not widget.is_read_only()
    ):
        return widget
    else:
        return None
    
def get_selected_line(text_widget: EditorCodeViewText, only_lineno=True) -> int | tuple[int, int]:
    """
    Get the number of the selected line in the text editor. If only_lineno is True, 
    get only the line number.
    
    Note: Before using this method you should check if several lines are selected
    by invoking the method `assert_one_line_is_selected()` located in this file.
    
    Args:
        text_widget (Widget): The text selected in the text editor. 
                    The value of this parameter must be the result 
                    of invoking the get_focused_writable_text() method. 
        only_lineno (bool): If True, only the number of the selected line is returned.
    Returns:
        int | tuple[int, int]: Returns (lineno, column). If `only_lineno` is True, 
        returns only lineno.
    """
    # A text is selected in the editor => can't tell the exact line of the test to run
    lineno, column = map(int, text_widget.index(tk.INSERT).split("."))
    return lineno if only_lineno else (lineno, column)

def assert_one_line_is_selected() -> bool:
    """
    Returns True if only one line is selected, otherwise an exception is raised.
    """
    text = get_focused_writable_text()
    if text :
        return len(text.tag_ranges("sel")) == 0
    return False
        
def add_random_suffix(word):
    """Add a random suffix to the given word.
     
    The suffix is added in the following format 'word_suffix'. An underscore separates the two words.
    The suffix is assumed to be long(more than 9 caracters).
    
    Args:
        word (str): A word.

    Returns:
        str: Returns the given word with a random suffix appended after an underscore.
    """
    import string
    from random import shuffle, randint
    divider = "_"
    alphabet = list(string.ascii_lowercase + string.ascii_uppercase)
    
    # Divide by three to avoid the first index being so close to the last index.
    # The first index should be smaller than the last index, so we can have a long suffix.
    first_index = randint(0, len(alphabet)//3) 
    last_index = randint(len(alphabet)//2, len(alphabet)-1)
    
    shuffle(alphabet) 
    suffix = "".join(alphabet[first_index: last_index])
    return word + divider + suffix

def get_font_family_option(option="editor_font_family") -> str:
    """Retrieves the value of the "font family" option from the "options" menu.

    Returns:
        str: The value of the setted font family.
    """
    option = option if option else "editor_font_family"
    return thonny.get_workbench().get_option("view.%s" % option)

def get_font_size_option():
    """Retrieves the value of the "font size" option from the "options" menu.

    Returns:
        int: The value of the setted font size.
    """
    workbench = thonny.get_workbench()
    return workbench._guard_font_size(workbench.get_option("view.editor_font_size"))

def get_image_path(basename:str):
    """Get the absolute path to the image in the /res directory

    Args:
        basename (str): just the name of the file inluding it's extension. 
        For example: "icon.png"

    Returns:
        str: the absolute path to the image located in the docs/res directory
    """
    return os.path.join(os.path.abspath(os.path.dirname(__file__)), "docs", "res", basename)

def get_photoImage(image_name:str):
    """Returns a PhotoImage object from the given image path.

    Args:
        image_name (str): The basename of the image with its extension. 
        The image must be located in the "docs/res" directory.   
    Returns:
        PhotoImage: A PhotoImage object.
    """
    # le chemin c'est par rapport au dossier du ce fichier 
    # donc il faut remonter vers /thonnycontrib/ pour rentrer dans docs/res
    icon_path = get_image_path(image_name)
    return tk.PhotoImage(name=get_basename(icon_path), file=icon_path)

def send_current_state(state:str, command_name=BACKEND_COMMAND, **options):
    """
    Sends the given state with given arg to the front(L1TestRunner).

    Args:
        state (str): The current state to send.
        command_name (str, optional): The name of the command. Default is BACKEND_COMMAND.
        **options: Other options to include in the message.

    Raises:
        AssertionError: If the backend is not initialized.
    """
    from thonny.common import InlineResponse
    import thonnycontrib.backend.l1test_backend
    backend = thonnycontrib.backend.l1test_backend.BACKEND
    assert backend != None
    backend.send_message(InlineResponse(command_name=command_name, state=state, **options))

def clear_env_vars(*vars):
    """
    Clears the environnement variables.
    """
    for var in vars:
        os.environ.pop(var, None)


# Auteur : Esteban COLLARD, Nordine EL AMMARI

from .ExampleVerdict import ExampleVerdict

class ExceptionVerdict(ExampleVerdict):
    def __init__(self, filename, lineno, tested_line, expected_result, error_details):
        super().__init__(filename, lineno, tested_line, expected_result, error_details)

    def get_color(self):
        return "red"

    def isSuccess(self):
        return False
    
    def _verdict_message(self):
        return "Test raises exception"


from typing import List
import re
from thonny import get_workbench
from functools import partial
from ..properties import PLUGIN_NAME
import tkinter as tk
from ..utils import  get_photoImage 
from thonnycontrib import l1test_frontend 

_OUTLINER_REGEX = r"\s*(?P<type>def|class)[ ]+(?P<name>[\w]+)"

class OutlinedNode():
    """
    This class represents an outlined node. An outlined node is either a class or a function.
    """ 
    def __init__(self, type:str, name:str, lineno:int) -> None:
        self.__type = type
        self.__name = name
        self.__lineno = lineno
        self.__image = get_photoImage("outline_class.png" if self.__type == "class" else "outline_method.gif")
    
    def get_type(self):
        return self.__type  
    
    def get_name(self):
        return self.__name  
    
    def get_lineno(self):
        return self.__lineno  
    
    def get_image(self):
        return self.__image 
    
    def __str__(self) -> str:
        return "{type: %s, name: %s, line: %s" % (self.__type, self.__name, self.__lineno)   

class L1TestOutliner():   
    """
    This class is responsible for parsing the source code and building the menu which contains
    the outlined nodes (classes and functions). 
    
    The parsing is done using a regular expression. 
    No AST is used, because the parsing is done on the fly, when the user clicks on the menu.
    
    The regular expression is defined in the global variable `_OUTLINER_REGEX`.
    """
    def __init__(self) -> None:
        l1test_frontend._outliner = self # singleton        
        get_workbench()._init_menu()
        # create the menu with a postcommand which will be called every time the menu is clicked
        self.menu = tk.Menu(get_workbench().get_menu(PLUGIN_NAME), postcommand=self.update_menu)
      
    def __parse(self, source:str) -> List[OutlinedNode]:
        """
        Parses a source and returns a list of the outlined nodes. 
        The outlined nodes are either a class or a function. For 
        each outlined node an object of type `OutlinedNode` is built 
        in which we store the type (class/function), the name and the lineno
        of the outlined node.
        """
        lineno = 0
        for line in source.splitlines():
            lineno += 1
            match = re.match(_OUTLINER_REGEX, line) 
            if match:
                yield OutlinedNode(match.group("type"), match.group("name"), lineno)
    
    def from_source_post_menu(self, source):
        self.clear_menu()
        for outlined in self.__parse(source):
            label = "%s %s" % (outlined.get_type(), outlined.get_name())
            image = outlined.get_image()
            self.menu.add_command(label=label, 
                                  image=image,
                                  command=partial(run_tests_for_outlined_node, outlined.get_lineno()),
                                  activebackground="white",
                                  activeforeground="darkblue",
                                  compound=tk.LEFT)
            # should save a reference to the image otherwise it will be garbage collected
            setattr(self, image.name, image)  

    def update_menu(self):
        editor = get_workbench().get_editor_notebook().get_current_editor()
        if not editor:
            self.clear_menu()
            return
        source = editor.get_code_view().get_content()
        self.from_source_post_menu(source)
    
    def clear_menu(self):
        self.menu.delete(0, tk.END)
        
    def get_menu(self):
        return self.menu

def run_tests_for_outlined_node(lineno:int):
    """
    Cette fonction est invoquée quand un item (méthode) du menu `Run test for ...` est cliqué.
    Cette fonction permet d'envoyer au l1test_backend la commande L1test avec en argument
    is_selected=True.
    """
    from thonnycontrib.l1test_frontend import get_l1test_runner
    get_l1test_runner().run_l1test(selected_line=lineno)


# Auteur : Esteban COLLARD, Nordine EL AMMARI
# Modifications : Reda ID TALEB & Manal LAGHMICH

from setuptools import setup
import os.path

import thonnycontrib

thonnycontrib_path = os.path.dirname(os.path.abspath(thonnycontrib.__file__))
def get_packages_under_thonnycontrib(directory:str, namespace:str):
    packages = []
    for root, dirs, files in os.walk(directory):
        if "__init__.py" in files:
            splitted: list[str] = root.split(os.sep)            
            try:
                index = splitted.index(namespace)
                sub_module = ".".join(splitted[index:])
                packages.append(sub_module)
            except:pass
    return packages

packages = get_packages_under_thonnycontrib(thonnycontrib_path, "thonnycontrib")

def get_packages_data(packages: list[str]=packages):
    py_packs = dict([(p, ["*.py"]) for p in packages if not p.endswith("docs")])
    other_packs = dict([(p, ["res/*"]) for p in packages if p.endswith("docs")])
    
    return py_packs | other_packs

setupdir = os.path.dirname(__file__)


setup(
    name="bimbam",
    version="4.6",
    author="idtaleb",
    description="A plug-in which adds a test framework",
    long_description="""A plug-in for Thonny which allows you to test your doc examples
 
More info: https://gitlab.univ-lille.fr/mirabelle.nebut/thonny-tests""",
    url="https://gitlab.univ-lille.fr/mirabelle.nebut/thonny-tests",
#    keywords="IDE education programming tests in documentation",
    classifiers=[
        "Topic :: Education :: Testing",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Education"
        ],
    platforms=["Windows", "macOS", "Linux"],
    python_requires=">=3.9",
    package_data=get_packages_data(),
    install_requires=["thonny>=4.0.0"],
    packages=packages,
)




from .l1test_runner import L1TestRunner
from .l1test_outliner_menu import L1TestOutliner

# Ceci est une implémentation de création d'un singleton pour L1TestRunner
_l1test_runner: L1TestRunner = None

def get_l1test_runner() -> L1TestRunner:
    """
    If there's no `L1TestRunner` instance creates one and returns it, 
    otherwise returns the current `L1TestRunner` instance.
    """
    return L1TestRunner() if not _l1test_runner else _l1test_runner


# Ceci est une implémentation de création d'un singleton pour OutlineParser
_outliner: L1TestOutliner = None

def get_outliner() -> L1TestOutliner:
    """
    Retourne une instance de `OutlineParser` en tant que singleton.
    """
    return L1TestOutliner() if not _outliner else _outliner


import os
from types import ModuleType

from .test_finder import FindSelectedL1DoctestStrategy
from ..exceptions import BackendException
from ..environement_vars import *
from .evaluator import Evaluator 
from thonny.common import ToplevelCommand, ToplevelResponse, InlineCommand
from ..properties import BACKEND_COMMAND, L1TEST_EXCEPTION, VERDICTS
from thonny.plugins.cpython_backend.cp_back import (
    Executor,
    MainCPythonBackend
)
import thonny.plugins.cpython_backend.cp_back
import pickle

BACKEND: MainCPythonBackend = thonny.plugins.cpython_backend.cp_back.get_backend()
 
class L1TestExecutor(Executor):
    def execute_source(self, source:str, filename:str, mode:str, ast_postprocessors):  
        """Cette fonction est invoquée par la méthode `BACKEND._execute_file(cmd, L1TestExecutor)`
        située en haut de ce fichier.

        Returns:
            dict: doit retourner, obligatoirement, un dictionnaire dont les données sont séralisées.
        """
        assert mode == "exec"
        try:
            evaluator = Evaluator(filename, source)
            selected_line: int = os.environ.get(SELECTED_LINE_VAR)
            if selected_line: # si un seul test a été ciblé
                selected_line = int(selected_line)
                evaluator.set_finder_strategy(FindSelectedL1DoctestStrategy(selected_line))
            
            verdicts = evaluator.evaluate()
            
            # on récupère la valeur de l'option qui indique si on doit importer le module dans le shell
            import_module_option = True if os.environ.get(IMPORT_MODULE_VAR) == "True" else False     
            # importation du module dans le shell.
            if import_module_option: 
                # importation du module dans le shell.
                self._import_module_in_shell(evaluator.get_module())                                                

            # Sérialiser les données en binaires. 
            serialized_data = pickle.dumps(verdicts)
            # The serialized data should necessary be a dictionary and the returned statement should
            # also be a dictionary
            return {VERDICTS: serialized_data}
        except BaseException as e:
            # pas besoin de sérialiser la valeur du dictionaire car il conteint que des types primitifs.
            return {L1TEST_EXCEPTION: self._build_backend_exception(e, str(e))}

    def _import_module_in_shell(self, module:ModuleType):
        """
        Imports the given `module` into Thonny's shell.
        
        Args:
            module (ModuleType): the module to be imported into the shell.
        """
        import __main__ 
        __main__.__dict__.update(module.__dict__) 
             
    def _build_backend_exception(self, exception:BaseException, message:str):
        """
            Returns a representation of the exception raised by the l1test backend.
            The representation includes the type of the exception and it's message.
            
            Note : the representation should necessary be a dictionary.
        Args:
            exception (BaseException): The raised exception.
            message (str): The message of the exception to be shown in the view.

        Returns:
            dict: a dictionary that contains information about the raised exception.
        """
        return {
                "type_name": exception.__class__.__name__,
                "message": message, 
                "prefix": exception.get_prefix_message() if isinstance(exception, BackendException) \
                                                         else None
            }
 
def _cmd_l1test(cmd: ToplevelCommand) -> ToplevelResponse:   
    """
    Cette fonction est invoquée lorsque un événement de type `ToplevelCommand` (associé 
    à la commande L1test) est récupéré.
    
    Args: 
        cmd(ToplevelCommand): L'événement `ToplevelCommand` associé à la commande L1test.
        
    Returns:
        ToplevelResponse: un événement de type ToplveleResponse qui contiendra la réponse
        renvoyé par `L1TestExecutor.execute_source()`.
    """ 
    response: ToplevelResponse = BACKEND._execute_file(cmd, L1TestExecutor)
    return response  
   
def load_plugin(): 
    """
        Cette fonction est importante car il est appelée par Thonny à son intialisation.
        
        Cette fonction doit déclarer les commandes magiques et leurs handlers.
    """
    BACKEND.add_command(BACKEND_COMMAND, _cmd_l1test)



import ast
from types import ModuleType
from thonnycontrib.backend.doctest_parser import ExampleExceptionExpected
from thonnycontrib.backend.evaluator import Evaluator
from thonnycontrib.exceptions import *
from thonnycontrib.backend.verdicts.ExceptionVerdict import ExceptionVerdict
from thonnycontrib.backend.verdicts.FailedVerdict import *
from thonnycontrib.backend.verdicts.PassedVerdict import *
from thonnycontrib.backend.verdicts.FailedWhenExceptionExpectedVerdict import FailedWhenExceptionExpectedVerdict
from thonnycontrib.tests.fixtures.backend_mock import *
from thonnycontrib.backend.ast_parser import *

from thonnycontrib.backend.doctest_parser import *
import unittest as ut

# ########################################################
#    Tous les tests qui suivent utilisent l'invite `$$e`
# ########################################################
class TestEvaluator(ut.TestCase):
    def setUp(self):
        self.evaluator = Evaluator(filename="<string>")
        self.mock_backend = backend_patch.start()
        
    def tearDown(self) -> None:
        del self.evaluator   
        backend_patch.stop()
    
    """
    Ce test vérifie:
    1. Le type `ExampleWithExpected` est le type `Example` extrait 
    par le doctest parser.
    2. Si une exception est attendue et que la fonction executée lève 
    cette exception alors on aura un `PassedVerdict`. Enfin, le flag du 
    l1doctest est `PASSED_FLAG`.
    """
    def test_evaluate_when_exception_expected_and_is_raised(self): 
        fake_source = \
"""
def f(a, b):
    '''
    $$e f(-1, -2)
    Exception
    '''
    if a < 0 and b < 0:
        raise Exception("a et b doivent être positifs")
    return a + b
""" 
        
        self.evaluator.set_source(fake_source)
        self.evaluator.set_module(self.__build_module_from_source(fake_source))
        
        # ###########################################################
        # ------------- Vérification du type Example ----------------
        self._assert_example_types(ExampleExceptionExpected)
        
        # ###################################################
        # ------------- Vérification du verdict -------------
        l1doctests = self.evaluator.evaluate()
        
        # on assure qu'il existe un seul l1doctest avec ses examples
        self.assertTrue(len(l1doctests) == 1)
        l1doctest = l1doctests[0]
        self.assertTrue(l1doctest.is_evaluated()) # s'assure que le l1doctest a été évalué
        self.assertEqual(l1doctest.get_flag(), L1DocTestFlag.PASSED_FLAG)
        
        # on assure que le verdict de l'example est un passed verdict
        examples = l1doctest.get_examples()
        self.assertTrue(len(examples) == 1) # on assure qu'il existe un seul Example
        self.assertTrue(isinstance(l1doctest.get_verdict_from_example(examples[0]), PassedVerdict))   
    
    """
    Ce test vérifie:
    1. Le type `ExampleWithExpected` est le type `Example` extrait 
    par le doctest parser.
    2. Si une exception est attendue et que la fonction executée ne lève pas
    cette exception alors on aura un `FailedWhenExceptionExpectedTest`. Enfin,
    le flag du l1doctest est `FAILED_FLAG`.
    """     
    def test_evaluate_when_exception_expected_but_not_raised(self):
        fake_source = \
"""
def f(a, b):
    '''
    $$e f(-1, -2)
    Exception
    '''
    if a < 0 and b < 0:
        return None
    return a + b
""" 
        
        self.evaluator.set_source(fake_source)
        self.evaluator.set_module(self.__build_module_from_source(fake_source))
        
        # ###########################################################
        # ------------- Vérification du type Example ----------------
        self._assert_example_types(ExampleExceptionExpected)
        
        # ###################################################
        # ------------- Vérification du verdict -------------
        l1doctests = self.evaluator.evaluate()

        # on assure qu'il existe un seul l1doctest avec ses examples
        self.assertTrue(len(l1doctests) == 1)
        l1doctest = l1doctests[0]
        self.assertTrue(l1doctest.is_evaluated()) # s'assure que le l1doctest a été évalué
        self.assertEqual(l1doctest.get_flag(), L1DocTestFlag.FAILED_FLAG)
        
        # on assure que le verdict de l'example est en erreur
        examples = l1doctest.get_examples()
        self.assertTrue(len(examples) == 1) # on assure qu'il existe un seul type Example
        
        verdict = l1doctest.get_verdict_from_example(examples[0])
        self.assertTrue(isinstance(verdict, FailedWhenExceptionExpectedVerdict))   
        self.assertTrue("Exception was not raised by `f(-1, -2)`" in verdict.get_details()) 
    
    """
    Ce test vérifie:
    1. Le type `ExampleWithExpected` est le type `Example` extrait 
    par le doctest parser.
    2. Si l'exception attendue n'hérite pas(directement ou indirectement) de 
    la classe `BaseException` alors un verdict de type `ExceptionTest` 
    est renvoyé. Enfin, le flag du l1doctest est `FAILED_FLAG`.
    """        
    def test_evaluate_when_exception_expected_but_dont_inherit_from_base_exception(self):
        
        fake_source = \
"""
class MyException(): pass

def f(a, b):
    '''
    $$e f(-1, -2)
    MyException
    '''
    if a < 0 and b < 0:
        raise MyException()
    return a + b
""" 
        
        self.evaluator.set_source(fake_source)
        self.evaluator.set_module(self.__build_module_from_source(fake_source))
        
        # ###########################################################
        # ------------- Vérification du type Example ----------------
        # on assure qu'il existe deux l1doctests (la classe MyException() et la fonction)
        l1doctests = self.evaluator.get_test_finder().find_l1doctests()
        self.assertTrue(len(l1doctests) == 2)
        for l1doctest in l1doctests: 
            if l1doctest.get_type() == ast.FunctionDef.__name__:
                examples = l1doctest.get_examples()
                self.assertTrue(len(examples) == 1) # on assure qu'il existe un seul type Example 
                self.assertTrue(isinstance(examples[0], ExampleExceptionExpected))
        
        # ###################################################
        # ------------- Vérification du verdict -------------
        l1doctests = self.evaluator.evaluate()

        # on assure qu'il existe deux l1doctestz avec ses examples
        self.assertTrue(len(l1doctests) == 2) 
        l1doctest = l1doctests[1] # on récupère le l1doctest de la fonction
        self.assertEqual(l1doctest.get_type(), ast.FunctionDef.__name__)
        self.assertTrue(l1doctest.is_evaluated()) # s'assure que le l1doctest a été évalué
        self.assertEqual(l1doctest.get_flag(), L1DocTestFlag.FAILED_FLAG)
        
        # on assure que le verdict de l'example est en erreur
        examples = l1doctest.get_examples()
        self.assertTrue(len(examples) == 1) # on assure qu'il existe un seul type Example
        
        verdict = l1doctest.get_verdict_from_example(examples[0])
        self.assertTrue(isinstance(verdict, ExceptionVerdict))   
        self.assertTrue(TypeError.__name__ in verdict.get_details()) 
        self.assertTrue("is not a class of type exception." in verdict.get_details())
    
    """
    Ce test vérifie:
    1. Le type `ExampleWithExpected` est le type `Example` extrait 
    par le doctest parser.
    2. Si l'exception attendue n'est pas retrouvée(n'est jamais déclarée)
    alors le verdict `ExceptionVerdict` est renvoyé. Enfin, le flag du
    l1doctest est `FAILED_FLAG`.
    """
    def test_evaluate_when_exception_expected_but_is_not_declared(self):    
        fake_source = \
"""
def f(a, b):
    '''
    $$e f(-1, -2)
    UnknownException
    '''
    if a < 0 and b < 0:
        raise Exception()
    return a + b
""" 
        self.evaluator.set_source(fake_source)
        self.evaluator.set_module(self.__build_module_from_source(fake_source))
        
        # ###########################################################
        # ------------- Vérification du type Example ----------------
        # on assure qu'il existe seul noeud AST
        self._assert_example_types(ExampleExceptionExpected)
        
        # ###################################################
        # ------------- Vérification du verdict -------------        
        l1doctests = self.evaluator.evaluate()

        # on assure qu'il existe un seul l1doctest avec ses examples
        self.assertTrue(len(l1doctests) == 1) 
        l1doctest = l1doctests[0] # on récupère le l1doctest de la fonction
        self.assertTrue(l1doctest.is_evaluated()) # s'assure que le l1doctest a été évalué
        self.assertEqual(l1doctest.get_flag(), L1DocTestFlag.FAILED_FLAG)
        
        # on assure que le verdict de l'example est en erreur
        examples = l1doctest.get_examples()
        self.assertTrue(len(examples) == 1) # on assure qu'il existe un seul type Example
        
        verdict = l1doctest.get_verdict_from_example(examples[0])
        self.assertTrue(isinstance(verdict, ExceptionVerdict))   
        self.assertTrue(NameError.__name__ in verdict.get_details()) 
        self.assertTrue("The expected exception `UnknownException` cannot be found." in verdict.get_details())
    
    """
    Ce test vérifie:
    1. Le type `ExampleWithExpected` est le type `Example` extrait 
    par le doctest parser.
    2. Si l'exception attendue n'est pas levée par la fonction executée
    alors le verdict `FailedWhenExceptionExpectedVerdict` est renvoyé. Enfin,
    le flag du l1doctest est `FAILED_FLAG`.
    """        
    def test_evaluate_when_exception_expected_but_an_other_exception_is_raised(self): 
        fake_source = \
"""
def f(a, b):
    '''
    $$e f(-1, -2)
    Exception
    '''
    if a < 0 and b < 0:
        raise ValueError()
    return a + b
""" 
        self.evaluator.set_source(fake_source)
        self.evaluator.set_module(self.__build_module_from_source(fake_source))
        
        # ###########################################################
        # ------------- Vérification du type Example ----------------
        # on assure qu'il existe seul noeud AST
        self._assert_example_types(ExampleExceptionExpected)
        
        # ###################################################
        # ------------- Vérification du verdict -------------
        l1doctests = self.evaluator.evaluate()

        # on assure qu'il existe un seul l1doctest avec ses examples
        self.assertTrue(len(l1doctests) == 1) 
        l1doctest = l1doctests[0] # on récupère le l1doctest de la fonction
        self.assertTrue(l1doctest.is_evaluated()) # s'assure que le l1doctest a été évalué
        self.assertEqual(l1doctest.get_flag(), L1DocTestFlag.FAILED_FLAG)
        
        # on assure que le verdict de l'example est en erreur
        examples = l1doctest.get_examples()
        self.assertTrue(len(examples) == 1) # on assure qu'il existe un seul type Example
        
        verdict = l1doctest.get_verdict_from_example(examples[0])
        self.assertTrue(isinstance(verdict, FailedWhenExceptionExpectedVerdict))   
        self.assertTrue(ValueError.__name__ in verdict.get_details()) 
        self.assertTrue("Exception was not raised by `f(-1, -2)`\nInstead, it raises :" in verdict.get_details())
    
    """
    Ce test vérifie:
    1. Le type `ExampleWithExpected` est le type `Example` extrait 
    par le doctest parser.
    2. Si le test contient une syntaxe erreur alors un verdict `ExceptionVerdict` 
    est renvoyé.
    """       
    def test_evaluate_when_syntax_error(self):
        fake_source = \
"""
def f(a, b):
    '''
    $$e f(-1, -2
    Exception
    '''
    if a < 0 and b < 0:
        raise ValueError()
    return a + b
""" 
        self.evaluator.set_source(fake_source)
        self.evaluator.set_module(self.__build_module_from_source(fake_source))
        
        # ###########################################################
        # ------------- Vérification du type Example ----------------
        # on assure qu'il existe seul noeud AST
        self._assert_example_types(ExampleExceptionExpected)
        
        # ###################################################
        # ------------- Vérification du verdict -------------
        l1doctests = self.evaluator.evaluate()

        # on assure qu'il existe un seul l1doctest avec ses examples
        self.assertTrue(len(l1doctests) == 1) 
        l1doctest = l1doctests[0] # on récupère le l1doctest de la fonction
        self.assertTrue(l1doctest.is_evaluated()) # s'assure que le l1doctest a été évalué
        self.assertEqual(l1doctest.get_flag(), L1DocTestFlag.FAILED_FLAG)
        
        # on assure que le verdict de l'example est en erreur
        examples = l1doctest.get_examples()
        self.assertTrue(len(examples) == 1) # on assure qu'il existe un seul type Example
        
        verdict = l1doctest.get_verdict_from_example(examples[0])
        self.assertTrue(isinstance(verdict, ExceptionVerdict))   
        self.assertTrue(SyntaxError.__name__ in verdict.get_details()) 
    
    """
    Ce test vérifie:
    1. Le type `ExampleWithExpected` est le type `Example` extrait 
    par le doctest parser.
    2. Si le want est un string vide ou un string contenant que des espaces
    alors le verdict `ExceptionVerdict` est renvoyé. Enfin, le flag du
    l1doctest est `FAILED_FLAG`.
    """         
    def test_evaluate_when_exception_is_blank_or_empty(self):
        fake_source = \
"""
def f(a, b):
    '''
    $$e f(-1, -2)   # Example of an empty <want> 
    $$e f(-1, -2)   # Example of a blank <want>
    
    '''
    if a < 0 and b < 0:
        raise Exception()
    return a + b
""" 
        self.evaluator.set_source(fake_source)
        self.evaluator.set_module(self.__build_module_from_source(fake_source))
        
        # ###########################################################
        # ------------- Vérification du type Example ----------------
        self._assert_example_types(ExampleExceptionExpected, 2)
        
        # ###################################################
        # ------------- Vérification du verdict -------------
        l1doctests = self.evaluator.evaluate()

        # on assure qu'il existe un seul l1doctest avec ses examples
        self.assertTrue(len(l1doctests) == 1) 
        l1doctest = l1doctests[0] # on récupère le l1doctest de la fonction
        self.assertTrue(l1doctest.is_evaluated()) # s'assure que le l1doctest a été évalué
        self.assertEqual(l1doctest.get_flag(), L1DocTestFlag.FAILED_FLAG)
        
        # on assure que le verdict de l'example est en erreur
        examples = l1doctest.get_examples()
        self.assertTrue(len(examples) == 2) # on assure qu'il existe un seul type Example
        for example in examples:   
            verdict = example.get_verdict()
            self.assertTrue(isinstance(verdict, ExceptionVerdict))   
            self.assertTrue(ValueError.__name__ in verdict.get_details())  
            self.assertTrue("The expected exception cannot be empty or blank" in verdict.get_details())
    
    
    def _assert_example_types(self, example_type: Example, len_examples=1):
        """
            Assert that the expected `Example` type is of type the given `example_type`.
            Also assert the number of the retrieved examples by given the `len_examples`
            parameter.
        """
        l1doctests = self.evaluator.get_test_finder().find_l1doctests()
        
        # on assure qu'il existe un seul noeud AST
        self.assertTrue(len(l1doctests) == 1)
        for l1_docTest in l1doctests:
            examples = l1_docTest.get_examples()
            self.assertTrue(len(examples) == len_examples) # on assure qu'il existe un seul type Example 
            for i in range(len_examples):
                self.assertTrue(isinstance(examples[i], example_type))
    
    def __build_module_from_source(self, source: str) -> ModuleType:
        """
        Build a module containing the functions declared in the given `source`.
        """
        from types import ModuleType
        fake_module = ModuleType(self.evaluator.get_filename())
        exec(source, fake_module.__dict__)
        return fake_module


if __name__ == '__main__':
    ut.main(verbosity=2)   
        


from ast import AST
import unittest as ut

from thonnycontrib.backend.evaluator import L1TestFinder
from thonnycontrib.backend.doctest_parser import (
    ExampleWithoutExpected, 
    ExampleExceptionExpected, 
    ExampleWithExpected
)
from thonnycontrib.exceptions import *
from thonnycontrib.backend.doctest_parser import *

# ########################################################
#  Tous les tests qui suivent testent le phase du parsing 
# ########################################################
class TestTestFinder(ut.TestCase):
    def setUp(self):
        self.test_finder = L1TestFinder(filename="<string>")
    
    """
    Ce test vérifie que s'il n'y a aucun test dans le noeud ast
    alors aucun type `Example` n'est renvoyé mais plutôt une liste vide.
    """
    def test_doctest_parser_when_no_examples(self):
        fake_source = \
"""
def f(a, b):
    if a < 0 and b < 0:
        return None
    return a + b
"""
        
        self.test_finder.set_source(fake_source)
        l1doctests = self.test_finder.find_l1doctests()

        # on assure qu'il existe un seul noeud AST
        self.assertTrue(len(l1doctests) == 1)
        for l1doctest in l1doctests:
            # on assure que le noeud AST ne contient aucun type Example
            self.assertEqual(l1doctest.get_examples(), []) 
    
    """
    Ce test vérifie que si l'invite est `$$$` et qu'il n'existe pas
    un `want` alors on aura le type `ExampleWithoutExpected`.
    """    
    def test_doctest_parser_when_ExampleWithoutExpected(self):
        fake_source = \
"""
def f(a, b):
    '''
    $$$ a=0
    
    '''
    if a < 0 and b < 0:
        return None
    return a + b
"""
        self.test_finder.set_source(fake_source)
        l1doctests = self.test_finder.find_l1doctests()
        
        # on assure qu'il existe un seul noeud ast
        self.assertTrue(len(l1doctests) == 1)
        
        for l1doctest in l1doctests:
            examples = l1doctest.get_examples()
            self.assertTrue(len(examples) == 1) # on assure qu'il existe un seul type Example 
            self.assertTrue(isinstance(examples[0], ExampleWithoutExpected))
    
    """
    Ce test vérifie que si l'invite est `$$$` et qu'il existe
    un `want` alors on aura le type `ExampleWithExpected`.        
    """
    def test_doctest_parser_when_ExampleWithExpected(self):
        
        fake_source = \
"""
def f(a, b):
    '''
    $$$ f(2, 3)
    5
    '''
    if a < 0 and b < 0:
        return None
    return a + b
"""
        self.test_finder.set_source(fake_source)
        l1doctests = self.test_finder.find_l1doctests()
        
        # on assure qu'il existe un seul noeud ast
        self.assertTrue(len(l1doctests) == 1)
        
        for l1doctest in l1doctests:
            examples = l1doctest.get_examples()
            self.assertTrue(len(examples) == 1) # on assure qu'il existe un seul type Example 
            self.assertTrue(isinstance(examples[0], ExampleWithExpected))
     
    """
    Ce test vérifie que si l'invite est `$$e` alors on aura 
    le type `ExampleExceptionExpected`.
    """       
    def test_doctest_parser_when_ExampleExceptionExpected(self):
        
        fake_source = \
"""
def f(a, b):
    '''
    $$e f(2, 3)
    Exception
    '''
    if a < 0 and b < 0:
        raise Exception("a et b doivent être positifs")
    return a + b
"""
        self.test_finder.set_source(fake_source)
        l1doctests = self.test_finder.find_l1doctests()
        
        # on assure qu'il existe un seul noeud ast
        self.assertTrue(len(l1doctests) == 1)
        
        for l1doctest in l1doctests:
            examples = l1doctest.get_examples()
            self.assertTrue(len(examples) == 1) # on assure qu'il existe un seul type Example 
            self.assertTrue(isinstance(examples[0], ExampleExceptionExpected))
    
    """
    Ce test vérifie que si une exception est attendue alors que
    l'invite de commande n'est pas un `$$e`, on aura pas alors le type 
    `ExampleExceptionExpected`.
    """
    def test_doctest_parser_when_unauthorized_exception(self):
        fake_source = \
"""
def f(a, b):
    '''
    $$$ f(-2, -3)
    Exception
    '''
    if a < 0 and b < 0:
        raise Exception("a et b doivent être positifs")
    return a + b
"""
        self.test_finder.set_source(fake_source)
        l1doctests = self.test_finder.find_l1doctests()
        
        # on assure qu'il existe un seul noeud ast
        self.assertTrue(len(l1doctests) == 1)
        
        for l1doctest in l1doctests:
            examples = l1doctest.get_examples()
            self.assertTrue(len(examples) == 1) # on assure qu'il existe un seul type Example 
            # on assure que ce n'est pas un type `ExampleExceptionExpected`
            self.assertTrue(not isinstance(examples[0], ExampleExceptionExpected))
            # mais que c'est un type `ExampleWithExpected`
            self.assertTrue(isinstance(examples[0], ExampleWithExpected))

    """
    Ce test vérifie:
    1. Quand il manque un espace après l'invite `$$$` alors 
        une exception de type `SpaceMissingAfterPromptException` est levée
    """
    def test_evaluate_when_a_space_is_missing_after_prompt_case_1(self):
        fake_source = \
"""
def f(a, b):
    '''
    $$$f(1, 1)
    2
    '''
    if a < 0 and b < 0:
        return None
    return a + b
""" 
        
        self.test_finder.set_source(fake_source)
        with self.assertRaises(SpaceMissingAfterPromptException):
            self.test_finder.find_l1doctests()
    
    """
    Ce test vérifie:
    1. Quand il manque un espace après l'invite `$$e` alors 
        une exception de type `SpaceMissingAfterPromptException` est levée
    """        
    def test_evaluate_when_a_space_is_missing_after_prompt_case_2(self): 
        fake_source = \
"""
def f(a, b):
    '''
    $$ef(1, 1)
    2
    '''
    if a < 0 and b < 0:
        return None
    return a + b
""" 
        
        self.test_finder.set_source(fake_source)
        with self.assertRaises(SpaceMissingAfterPromptException):
            self.test_finder.find_l1doctests()

if __name__ == '__main__':
    ut.main(verbosity=2)   